"""Calculate the n-th discrete difference along the given axis."""

from __future__ import annotations
from typing import Optional

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.diff)
def diff(
    a: PolyLike,
    n: int = 1,
    axis: int = -1,
    prepend: Optional[PolyLike] = None,
    append: Optional[PolyLike] = None,
) -> ndpoly:
    """
    Calculate the n-th discrete difference along the given axis.

    The first difference is given by ``out[i] = a[i+1] - a[i]`` along the given
    axis, higher differences are calculated by using `diff` recursively.

    Args:
        a:
            Input array
        n:
            The number of times values are "differenced". If zero, the input is
            returned as-is.
        axis:
            The axis along which the difference is taken, default is the last
            axis.
        prepend, append:
            Values to prepend or append to `a` along axis prior to
            performing the difference. Scalar values are expanded to
            arrays with length 1 in the direction of axis and the shape
            of the input array in along all other axes.  Otherwise the
            dimension and shape must match `a` except along axis.

    Return:
        The n-th differences. The shape of the output is the same as `a` except
        along `axis` where the dimension is smaller by `n`. The type of the
        output is the same as the type of the difference between any two
        elements of `a`. This is the same as the type of `a` in most cases.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([1, q0, q1, q0**2, q1-1])
        >>> numpoly.diff(poly)
        polynomial([q0-1, q1-q0, q0**2-q1, -q0**2+q1-1])
        >>> numpoly.diff(poly, n=2)
        polynomial([q1-2*q0+1, q0**2-2*q1+q0, -2*q0**2+2*q1-1])
        >>> poly = numpoly.polynomial([[q0, 1], [2, q1]])
        >>> numpoly.diff(poly)
        polynomial([[-q0+1],
                    [q1-2]])
        >>> numpoly.diff(poly, prepend=7, append=q1)
        polynomial([[q0-7, -q0+1, q1-1],
                    [-5, q1-2, 0]])

    """
    if append is not None:
        if prepend is not None:
            a, append, prepend = numpoly.align_exponents(a, append, prepend)
        else:
            a, append = numpoly.align_exponents(a, append)
    elif prepend is not None:
        a, prepend = numpoly.align_exponents(a, prepend)
    else:
        a = numpoly.aspolynomial(a)

    out = None
    for key in a.keys:
        kwargs = {}
        if append is not None:
            kwargs["append"] = append.values[key]
        if prepend is not None:
            kwargs["prepend"] = prepend.values[key]
        tmp = numpy.diff(a.values[key], n=n, axis=axis, **kwargs)

        if out is None:
            out = numpoly.ndpoly(
                exponents=a.exponents,
                shape=tmp.shape,
                names=a.indeterminants,
                dtype=tmp.dtype,
            )
        out.values[key] = tmp
    assert out is not None
    return numpoly.clean_attributes(out)



"""Save a polynomial array to a text file."""

from __future__ import annotations
from typing import Optional
from os import PathLike

import numpy
from numpy.lib.recfunctions import structured_to_unstructured
import numpoly

from ..baseclass import PolyLike
from ..dispatch import implements


HEADER_TEMPLATE = "numpoly:{version} names:{names} keys:{keys} shape:{shape}"


@implements(numpy.savetxt)
def savetxt(
    fname: PathLike,
    X: PolyLike,
    fmt: str = "%.18e",
    delimiter: str = " ",
    newline: str = "\n",
    header: str = "",
    footer: str = "",
    comments: str = "# ",
    encoding: Optional[str] = None,
):
    """
    Save a polynomial array to a text file.

    To store a polynomial to text string, the polynomial is converted through
    a flat structured array, to a matrix. Extra meta information about the
    indeterminant names, exponent keys and array shape are all stored
    separate at the top of the header of the output file.

    Args:
        fname:
            If the filename ends in ``.gz``, the file is automatically saved in
            compressed gzip format. `loadtxt` understands gzipped files
            transparently.
        X:
            Data to be saved to a text file.
        fmt:
            A single format (%10.5f), a sequence of formats, or a multi-format
            string, e.g. 'Iteration %d -- %10.5f', in which case `delimiter` is
            ignored. For complex `X`, the legal options for `fmt` are:

            * A single specifier, `fmt='%.4e'`, resulting in numbers formatted
              like `' (%s+%sj)' % (fmt, fmt)`.
            * A full string specifying every real and imaginary part, e.g.
              `' %.4e %+.4ej %.4e %+.4ej %.4e %+.4ej'` for 3 columns.
            * A list of specifiers, one per column - in this case, the real
              and imaginary part must have separate specifiers,
              e.g. `['%.3e + %.3ej', '(%.15e%+.15ej)']` for 2 columns.

        delimiter:
            String or character separating columns.
        newline:
            String or character separating lines.
        header:
            String that will be written at the beginning of the file.
        footer:
            String that will be written at the end of the file.
        comments:
            String that will be prepended to the ``header`` and ``footer``
            strings, to mark them as comments. Default: '# ',  as expected by
            e.g. ``numpoly.loadtxt``.
        encoding:
            Encoding used to encode the outputfile. Does not apply to output
            streams. If the encoding is something other than 'bytes' or
            'latin1' you will not be able to load the file in NumPy versions <
            1.14. Default is 'latin1'.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([1, q0, q1**2-1])
        >>> numpoly.savetxt("/tmp/poly.txt", poly)
        >>> numpoly.loadtxt("/tmp/poly.txt")
        polynomial([1.0, q0, q1**2-1.0])
        >>> numpoly.savetxt("/tmp/poly.txt", poly, header="my header")
        >>> numpoly.loadtxt("/tmp/poly.txt", skiprows=1)
        polynomial([1.0, q0, q1**2-1.0])

    """
    if isinstance(X, numpoly.ndpoly):
        numpoly_header = HEADER_TEMPLATE.format(
            version=numpoly.__version__,
            names=",".join(X.names),
            keys=",".join(X.keys),
            shape=",".join(str(idx) for idx in X.shape),
        )
        if header:
            header = numpoly_header + "\n" + header
        else:
            header = numpoly_header
        X = structured_to_unstructured(X.values.ravel())

    numpy.savetxt(
        fname=fname,
        X=X,
        fmt=fmt,
        delimiter=delimiter,
        newline=newline,
        header=header,
        footer=footer,
        comments=comments,
        encoding=encoding,
    )



"""Return a full array with the same shape and type as a given array."""

from __future__ import annotations
from typing import Optional, Sequence

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.full_like)
def full_like(
    a: PolyLike,
    fill_value: numpy.typing.ArrayLike,
    dtype: Optional[numpy.typing.DTypeLike] = None,
    order: str = "K",
    subok: bool = True,
    shape: Optional[Sequence[int]] = None,
) -> ndpoly:
    """
    Return a full array with the same shape and type as a given array.

    Args:
        a:
            The shape and data-type of `a` define these same attributes of
            the returned array.
        fill_value:
            Fill value. Must be broadcast compatible with shape.
        dtype:
            Overrides the data type of the result.
        order:
            Overrides the memory layout of the result. 'C' means C-order,
            'F' means F-order, 'A' means 'F' if `a` is Fortran contiguous,
            'C' otherwise. 'K' means match the layout of `a` as closely
            as possible.
        subok:
            If True, then the newly created array will use the sub-class
            type of 'a', otherwise it will be a base-class array. Defaults
            to True.
        shape:
            Overrides the shape of the result. If order='K' and the number of
            dimensions is unchanged, will try to keep order, otherwise,
            order='C' is implied.

    Return:
        Array of `fill_value` with the same shape and type as `a`.

    Example:
        >>> poly = numpoly.monomial(3)
        >>> poly
        polynomial([1, q0, q0**2])
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.full_like(poly, q1-1.)
        polynomial([q1-1, q1-1, q1-1])

    """
    del subok
    if not isinstance(a, numpy.ndarray):
        a = numpoly.polynomial(a)
    fill_value = numpoly.aspolynomial(fill_value)
    if shape is None:
        shape = a.shape
    if dtype is None:
        dtype = a.dtype
    if order in ("A", "K"):
        order = "F" if a.flags["F_CONTIGUOUS"] else "C"
    out = numpoly.ndpoly(
        exponents=fill_value.exponents,
        shape=tuple(shape),
        names=fill_value.indeterminants,
        dtype=dtype,
        order=order,
    )
    for key in fill_value.keys:
        out.values[key] = fill_value.values[key]
    return out



"""Return elements chosen from `x` or `y` depending on `condition`."""

from __future__ import annotations

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.where)
def where(condition: numpy.typing.ArrayLike, *args: PolyLike) -> ndpoly:
    """
    Return elements chosen from `x` or `y` depending on `condition`.

    .. note::
        When only `condition` is provided, this function is a shorthand for
        ``np.asarray(condition).nonzero()``. Using `nonzero` directly should be
        preferred, as it behaves correctly for subclasses. The rest of this
        documentation covers only the case where all three arguments a re
        provided.

    Args:
        condition:
            Where True, yield `x`, otherwise yield `y`.
        x:
            Values from which to choose. `x`, `y` and `condition` need to be
            broadcastable to some shape.

    Return:
        An array with elements from `x` where `condition` is True,
        and elements from `y` elsewhere.

    Example:
        >>> poly = numpoly.variable()*numpy.arange(4)
        >>> poly
        polynomial([0, q0, 2*q0, 3*q0])
        >>> numpoly.where([1, 0, 1, 0], 7, 2*poly)
        polynomial([7, 2*q0, 7, 6*q0])
        >>> numpoly.where(poly, 2*poly, 4)
        polynomial([4, 2*q0, 4*q0, 6*q0])
        >>> numpoly.where(poly)
        (array([1, 2, 3]),)

    """
    if isinstance(condition, numpoly.ndpoly):
        condition = numpy.any(numpy.asarray(condition.coefficients), 0).astype(bool)
    if not args:
        return numpy.where(condition)

    poly1, poly2 = numpoly.align_polynomials(*args)
    coefficients = [
        numpy.where(condition, x1, x2)
        for x1, x2 in zip(poly1.coefficients, poly2.coefficients)
    ]
    dtype = numpy.result_type(poly1.dtype, poly2.dtype)
    return numpoly.polynomial_from_attributes(
        exponents=poly1.exponents,
        coefficients=coefficients,
        names=poly1.names,
        dtype=dtype,
    )



"""Convert the input to an polynomial array."""

from __future__ import annotations
from typing import Optional, Union, Tuple

import numpy.typing
import numpoly

from ..baseclass import PolyLike, ndpoly


def aspolynomial(
    poly_like: PolyLike,
    names: Union[None, str, Tuple[str, ...], ndpoly] = None,
    dtype: Optional[numpy.typing.DTypeLike] = None,
) -> ndpoly:
    """
    Convert the input to an polynomial array.

    Args:
        poly_like:
            Input to be converted to a `numpoly.ndpoly` polynomial type.
        names:
            Name of the indeterminant variables. If possible to infer from
            ``poly_like``, this argument will be ignored.
        dtype:
            Data type used for the polynomial coefficients.

    Return:
        Array interpretation of `poly_like`. No copy is performed if the input
        is already an ndpoly with matching indeterminants names and dtype.

    Example:
        >>> q0 = numpoly.variable()
        >>> numpoly.polynomial(q0) is q0
        False
        >>> numpoly.aspolynomial(q0) is q0
        True

    """
    remain = False
    if isinstance(poly_like, numpoly.ndpoly):
        remain = dtype is None or dtype == poly_like.dtype
        if names is not None:
            if isinstance(names, numpoly.ndpoly):
                names_ = names.names
            elif isinstance(names, str):
                names_ = (names,)
            else:
                names_ = tuple(names)
            if len(names_) == 1 and len(poly_like.names) > 1:
                names_ = tuple(
                    f"{names_[0]}{idx}" for idx in range(len(poly_like.indeterminants))
                )
            remain &= names_ == poly_like.names

    if remain:
        return poly_like  # type: ignore
    return numpoly.polynomial(poly_like, names=names, dtype=dtype)



"""Test utility functions."""
from pytest import raises
import numpy
import numpoly


def test_glexindex():
    assert not numpoly.glexindex(0).size
    assert numpy.all(numpoly.glexindex(1) == [[0]])
    assert numpy.all(numpoly.glexindex(5) ==
                     [[0], [1], [2], [3], [4]])
    assert numpy.all(numpoly.glexindex(2, dimensions=2) ==
                     [[0, 0], [1, 0], [0, 1]])
    assert numpy.all(numpoly.glexindex(start=2, stop=3, dimensions=2) ==
                     [[2, 0], [1, 1], [0, 2]])
    assert numpy.all(numpoly.glexindex(start=2, stop=[3, 4], dimensions=2) ==
                     [[2, 0], [1, 1], [0, 2], [0, 3]])
    assert numpy.all(numpoly.glexindex(start=[2, 5], stop=[3, 6], dimensions=2) ==
                     [[2, 0], [1, 1], [1, 2], [0, 5]])
    assert numpy.all(numpoly.glexindex(start=2, stop=4, dimensions=2, cross_truncation=0) ==
                     [[2, 0], [3, 0], [0, 2], [0, 3]])
    assert numpy.all(numpoly.glexindex(start=2, stop=4, dimensions=2, cross_truncation=1) ==
                     [[2, 0], [3, 0], [1, 1], [2, 1], [0, 2], [1, 2], [0, 3]])
    assert numpy.all(numpoly.glexindex(start=2, stop=4, dimensions=2, cross_truncation=2) ==
                     [[2, 0], [3, 0], [1, 1], [2, 1], [0, 2], [1, 2], [2, 2], [0, 3]])
    assert numpy.all(numpoly.glexindex(start=0, stop=2, dimensions=3) ==
                     [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]])

    assert numpy.all(numpoly.glexindex(start=0, stop=[1, 1, 1]) == [0, 0, 0])


def test_glexsort():
    indices = numpy.array([[0, 0, 0, 1, 2, 1],
                           [1, 2, 0, 0, 0, 1]])
    assert numpy.all(indices.T[numpoly.glexsort(indices)].T ==
                     [[0, 1, 2, 0, 1, 0],
                      [0, 0, 0, 1, 1, 2]])
    assert numpy.all(indices.T[numpoly.glexsort(indices, graded=True)].T ==
                     [[0, 1, 0, 2, 1, 0],
                      [0, 0, 1, 0, 1, 2]])
    assert numpy.all(indices.T[numpoly.glexsort(indices, reverse=True)].T ==
                     [[0, 0, 0, 1, 1, 2],
                      [0, 1, 2, 0, 1, 0]])
    assert numpy.all(indices.T[numpoly.glexsort(indices, graded=True, reverse=True)].T ==
                     [[0, 0, 1, 0, 1, 2],
                      [0, 1, 0, 2, 1, 0]])

    indices = numpy.array([4, 5, 6, 3, 2, 1])
    assert numpy.all(numpoly.glexsort(indices) == [5, 4, 3, 0, 1, 2])
    indices = numpy.array([[4, 5, 6, 3, 2, 1]])
    assert numpy.all(numpoly.glexsort(indices) == [5, 4, 3, 0, 1, 2])


def test_cross_truncate():
    indices = numpy.array(numpy.mgrid[:10, :10]).reshape(2, -1).T

    assert not numpy.any(numpoly.cross_truncate(indices, -1, norm=0))
    assert numpy.all(indices[numpoly.cross_truncate(indices, 0, norm=0)].T ==
                     [[0], [0]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, 1, norm=0)].T ==
                     [[0, 0, 1], [0, 1, 0]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, 2, norm=0)].T ==
                     [[0, 0, 0, 1, 2], [0, 1, 2, 0, 0]])

    assert not numpy.any(numpoly.cross_truncate(indices, -1, norm=1))
    assert numpy.all(indices[numpoly.cross_truncate(indices, 0, norm=1)].T ==
                     [[0], [0]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, 1, norm=1)].T ==
                     [[0, 0, 1], [0, 1, 0]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, 2, norm=1)].T ==
                     [[0, 0, 0, 1, 1, 2], [0, 1, 2, 0, 1, 0]])

    assert not numpy.any(numpoly.cross_truncate(indices, -1, norm=100))
    assert numpy.all(indices[numpoly.cross_truncate(indices, 0, norm=100)].T ==
                     [[0], [0]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, 1, norm=100)].T ==
                     [[0, 0, 1], [0, 1, 0]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, 2, norm=100)].T ==
                     [[0, 0, 0, 1, 1, 1, 2, 2], [0, 1, 2, 0, 1, 2, 0, 1]])

    assert not numpy.any(numpoly.cross_truncate(indices, -1, norm=numpy.inf))
    assert numpy.all(indices[numpoly.cross_truncate(indices, 0, norm=numpy.inf)].T ==
                     [[0], [0]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, 1, norm=numpy.inf)].T ==
                     [[0, 0, 1, 1], [0, 1, 0, 1]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, 2, norm=numpy.inf)].T ==
                     [[0, 0, 0, 1, 1, 1, 2, 2, 2], [0, 1, 2, 0, 1, 2, 0, 1, 2]])

    indices = numpy.array(numpy.mgrid[:10, :10, :10]).reshape(3, -1).T
    assert not numpy.any(numpoly.cross_truncate(indices, -1, 1))
    assert numpy.all(indices[numpoly.cross_truncate(indices, 0, 1)].T == [0, 0, 0])
    assert numpy.all(indices[numpoly.cross_truncate(indices, 1, 1)].T ==
                     [[0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, [0, 0, 1], 1)].T ==
                     [[0, 0], [0, 0], [0, 1]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, [1, 1, 2], 1)].T ==
                     [[0, 0, 0, 0, 1], [0, 0, 0, 1, 0], [0, 1, 2, 0, 0]])
    assert numpy.all(indices[numpoly.cross_truncate(indices, [1, 2, 3], 1)].T ==
                     [[0, 0, 0, 0, 0, 0, 0, 1],
                      [0, 0, 0, 0, 1, 1, 2, 0],
                      [0, 1, 2, 3, 0, 1, 0, 0]])



"""Check if a polynomial is constant or not."""

from __future__ import annotations

import numpy
import numpoly

from ..baseclass import PolyLike


def isconstant(poly: PolyLike) -> bool:
    """
    Check if a polynomial is constant or not.

    Args:
        poly:
            polynomial to check if is constant or not.

    Return:
        True if polynomial has no indeterminants.

    Example:
        >>> q0 = numpoly.variable()
        >>> numpoly.isconstant(numpoly.polynomial([q0]))
        False
        >>> numpoly.isconstant(numpoly.polynomial([1]))
        True

    """
    poly = numpoly.aspolynomial(poly)
    for exponent, coefficient in zip(poly.exponents, poly.coefficients):
        if not numpy.any(exponent):
            continue
        if numpy.any(coefficient):
            return False
    return True



"""Return the product of array elements over a given axis."""

from __future__ import annotations
from typing import Any, Optional, Sequence, Union

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.prod)
def prod(
    a: PolyLike,
    axis: Union[None, int, Sequence[int]] = None,
    dtype: Optional[numpy.typing.DTypeLike] = None,
    out: Optional[ndpoly] = None,
    keepdims: bool = False,
    **kwargs: Any,
) -> ndpoly:
    """
    Return the product of array elements over a given axis.

    Args:
        a : array_like
            Input data.
        axis : None or int or tuple of ints, optional
            Axis or axes along which a product is performed.  The default,
            axis=None, will calculate the product of all the elements in the
            input array. If axis is negative it counts from the last to the
            first axis. If axis is a tuple of ints, a product is performed on
            all of the axes specified in the tuple instead of a single axis or
            all the axes as before.
        dtype : dtype, optional
            The type of the returned array, as well as of the accumulator in
            which the elements are multiplied.  The dtype of `a` is used by
            default unless `a` has an integer dtype of less precision than the
            default platform integer.  In that case, if `a` is signed then the
            platform integer is used while if `a` is unsigned then an unsigned
            integer of the same precision as the platform integer is used.
        out : ndarray, optional
            Alternative output array in which to place the result. It must have
            the same shape as the expected output, but the type of the output
            values will be cast if necessary.
        keepdims : bool, optional
            If this is set to True, the axes which are reduced are left in the
            result as dimensions with size one. With this option, the result
            will broadcast correctly against the input array.
        initial : scalar, optional
            The starting value for this product.
        where : array_like of bool, optional
            Elements to include in the product.

    Return:
        An array shaped as `a` but with the specified axis removed.
        Returns a reference to `out` if specified.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([[[1, q0, q0**2],
        ...                             [q0+q1, q1, q1]]])
        >>> numpoly.prod(poly)
        polynomial(q0**3*q1**3+q0**4*q1**2)
        >>> numpoly.prod(poly, keepdims=True)
        polynomial([[[q0**3*q1**3+q0**4*q1**2]]])
        >>> numpoly.prod(poly, axis=1)
        polynomial([[q1+q0, q0*q1, q0**2*q1]])
        >>> numpoly.prod(poly, axis=2, keepdims=True)
        polynomial([[[q0**3],
                     [q1**3+q0*q1**2]]])
        >>> numpoly.prod(poly, axis=[1, 2])
        polynomial([[[q0**3*q1**3+q0**4*q1**2]]])

    """
    a = numpoly.aspolynomial(a)
    assert out is None
    if keepdims:
        if axis is None:
            out = _prod(numpoly.reshape(a, -1), axis=0)
            out = numpoly.reshape(out, (1,) * len(a.shape))
            return out
        elif isinstance(axis, int):
            axis = [axis]

    if axis is None:
        out = _prod(numpoly.reshape(a, -1), axis=0)

    elif isinstance(axis, int):
        out = _prod(a, axis=axis)

    else:
        for idx in axis:
            a = _prod(a, axis=idx)
            a = a[(slice(None),) * idx + (numpy.newaxis,)]
        out = a

    return out


def _prod(a: ndpoly, axis: int) -> ndpoly:
    """
    Backend for the product function.

    Args:
        a:
            Input data.
        axis:
            The axis to take product over.

    Return:
        An array shaped as `a` but with the specified axis removed.

    """
    axis = axis + a.ndim if axis < 0 else axis
    assert a.ndim > axis, (a, axis)
    indices = (slice(None),) * axis
    out = a[indices + (0,)]
    for idx in range(1, a.shape[axis]):
        out = numpoly.multiply(out, a[indices + (idx,)])
    assert len(out.shape) + 1 == len(a.shape)
    return out



"""Add arguments element-wise."""

from __future__ import annotations
from typing import Any, Optional

import numpy
import numpy.typing

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements, simple_dispatch


@implements(numpy.add)
def add(
    x1: PolyLike,
    x2: PolyLike,
    out: Optional[ndpoly] = None,
    where: numpy.typing.ArrayLike = True,
    **kwargs: Any,
) -> ndpoly:
    """
    Add arguments element-wise.

    Args:
        x1, x2:
            The arrays to be added. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out (Optional[numpy.ndarray]) -> ndpoly:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where (Union[bool, numpy.ndarray]) -> ndpoly:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Return:
        The sum of `x1` and `x2`, element-wise. This is a scalar if both
        `x1` and `x2` are scalars.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.add(q0, 4)
        polynomial(q0+4)
        >>> poly1 = q0**numpy.arange(9).reshape((3, 3))
        >>> poly2 = q1**numpy.arange(3)
        >>> numpoly.add(poly1, poly2)
        polynomial([[2, q1+q0, q1**2+q0**2],
                    [q0**3+1, q0**4+q1, q0**5+q1**2],
                    [q0**6+1, q0**7+q1, q0**8+q1**2]])

    """
    return simple_dispatch(
        numpy_func=numpy.add,
        inputs=(x1, x2),
        out=None if out is None else (out,),
        where=where,
        **kwargs,
    )



import numpy
from numpoly import polynomial
import numpoly


X, Y = numpoly.variable(2)


def test_align_polynomials():
    poly1 = 5+3*Y+3*X
    poly2 = 4.
    poly1_, poly2_ = numpoly.align_polynomials(poly1, poly2)
    assert poly1 == poly1_
    assert poly2 == poly2_
    assert numpy.all(poly1_.exponents == poly2_.exponents)
    assert poly1_.exponents.shape[-1] == 2
    assert poly1_.shape == poly2_.shape

    X_, Y_ = numpoly.align_polynomials(X, Y)
    assert not X_.shape
    assert not Y_.shape
    assert X_ == X
    assert Y_ == Y



"""Multiply arguments element-wise."""

from __future__ import annotations
from typing import Any, Optional

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.multiply)
def multiply(
    x1: PolyLike,
    x2: PolyLike,
    out: Optional[ndpoly] = None,
    where: numpy.typing.ArrayLike = True,
    **kwargs: Any,
) -> ndpoly:
    """
    Multiply arguments element-wise.

    Args:
        x1, x2:
            Input arrays to be multiplied. If ``x1.shape != x2.shape``, they
            must be broadcastable to a common shape (which becomes the shape of
            the output).
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Return:
        The product of `x1` and `x2`, element-wise. This is a scalar if
        both `x1` and `x2` are scalars.

    Example:
        >>> poly = numpy.arange(9.0).reshape((3, 3))
        >>> q0q1q2 = numpoly.variable(3)
        >>> numpoly.multiply(poly, q0q1q2)
        polynomial([[0.0, q1, 2.0*q2],
                    [3.0*q0, 4.0*q1, 5.0*q2],
                    [6.0*q0, 7.0*q1, 8.0*q2]])

    """
    x1, x2 = numpoly.align_indeterminants(x1, x2)

    dtype = numpy.result_type(x1, x2)
    shape = numpy.broadcast_shapes(x1.shape, x2.shape)

    where = numpy.asarray(where)
    exponents = numpy.unique(
        numpy.tile(x1.exponents, (len(x2.exponents), 1))
        + numpy.repeat(x2.exponents, len(x1.exponents), 0),
        axis=0,
    )
    out_ = (
        numpoly.ndpoly(
            exponents=exponents,
            shape=shape,
            names=x1.indeterminants,
            dtype=dtype,
        )
        if out is None
        else out
    )

    #    seen = set()
    #    for expon1, coeff1 in zip(x1.exponents, x1.coefficients):
    #        for expon2, coeff2 in zip(x2.exponents, x2.coefficients):
    #            key = (expon1 + expon2 + x1.KEY_OFFSET).ravel()
    #            key = key.view(f"U{len(expon1)}").item()
    #            if key in seen:
    #                out_.values[key] += numpy.multiply(
    #                    coeff1, coeff2, where=where, **kwargs
    #                )
    #            else:
    #                numpy.multiply(
    #                    coeff1, coeff2, out=out_.values[key], where=where, **kwargs
    #                )
    #            seen.add(key)
    #
    #    if out is None:
    #        out_ = numpoly.clean_attributes(out_)

    numpoly.cmultiply(
        x1.exponents,
        x2.exponents,
        x1.coefficients,
        x2.coefficients,
        x1.KEY_OFFSET,
        out_.values.ravel(),
    )
    if out is None:
        out_ = numpoly.clean_attributes(out_)

    return out_



"""Apply a function to 1-D slices along the given axis."""

from __future__ import annotations
from functools import wraps
from typing import Any, Callable, List

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.apply_along_axis)
def apply_along_axis(
    func1d: Callable[[PolyLike], PolyLike],
    axis: int,
    arr: PolyLike,
    *args: Any,
    **kwargs: Any,
) -> ndpoly:
    """
    Apply a function to 1-D slices along the given axis.

    Execute `func1d(a, *args)` where `func1d` operates on 1-D arrays and `a` is
    a 1-D slice of `arr` along `axis`.

    This is equivalent to (but faster than) the following use of `ndindex` and
    `s_`, which sets each of ``ii``, ``jj``, and ``kk`` to a tuple of indices::

        Ni, Nk = a.shape[:axis], a.shape[axis+1:]
        for ii in ndindex(Ni):
            for kk in ndindex(Nk):
                f = func1d(arr[ii+s_[:,]+kk])
                Nj = f.shape
                for jj in ndindex(Nj):
                    out[ii+jj+kk] = f[jj]

    Equivalently, eliminating the inner loop, this can be expressed as::

        Ni, Nk = a.shape[:axis], a.shape[axis+1:]
        for ii in ndindex(Ni):
            for kk in ndindex(Nk):
                out[ii+s_[...,]+kk] = func1d(arr[ii+s_[:,]+kk])

    Args:
        func1d:
            This function should accept 1-D arrays. It is applied to 1-D slices
            of `arr` along the specified axis.
        axis:
            Axis along which `arr` is sliced.
        arr:
            Input array.
        args:
            Additional arguments to `func1d`.
        kwargs:
            Additional named arguments to `func1d`.

    Return:
        The output array. The shape of `out` is identical to the shape of
        `arr`, except along the `axis` dimension. This axis is removed, and
        replaced with new dimensions equal to the shape of the return value of
        `func1d`. So if `func1d` returns a scalar `out` will have one fewer
        dimensions than `arr`.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> b = numpoly.polynomial([[1, 2, 3*q0],
        ...                         [3, 6*q1, 6],
        ...                         [2, 7, 9]])
        >>> numpoly.apply_along_axis(numpoly.mean, 0, b)
        polynomial([2.0, 2.0*q1+3.0, q0+5.0])
        >>> numpoly.apply_along_axis(numpoly.mean, 1, b)
        polynomial([q0+1.0, 2.0*q1+3.0, 6.0])

    """
    collection: List[ndpoly] = list()

    @wraps(func1d)
    def wrapper_func(array):
        """Wrap func1d function."""
        # Align indeterminants in case slicing changed them
        array = numpoly.polynomial(
            array, names=arr.indeterminants, allocation=arr.allocation
        )
        array, _ = numpoly.align.align_indeterminants(array, arr.indeterminants)

        # Evaluate function
        out = func1d(array, *args, **kwargs)

        # Restore indeterminants in case func1d changed them.
        out, _ = numpoly.align.align_indeterminants(out, arr.indeterminants)

        # Return dummy index integer value that will be replaced with
        # polynomials afterwards.
        ret_val = len(collection) * numpy.ones(out.shape, dtype=int)
        collection.append(out)
        return ret_val

    # Initiate wrapper
    arr = numpoly.aspolynomial(arr)
    out = numpy.apply_along_axis(wrapper_func, axis=axis, arr=arr.values)

    # align exponents
    polynomials = numpoly.align.align_exponents(*collection)
    dtype = numpoly.result_type(*polynomials)

    # Store results into new array
    ret_val = numpoly.ndpoly(
        exponents=polynomials[0].exponents,
        shape=out.shape,
        names=polynomials[0].indeterminants,
        dtype=dtype,
    ).values
    for idx, polynomial in enumerate(polynomials):
        ret_val[out == idx] = polynomial.values

    return numpoly.polynomial(
        ret_val,
        dtype=dtype,
        names=polynomials[0].indeterminants,
        allocation=polynomials[0].allocation,
    )



"""View inputs as arrays with at least two dimensions."""

from __future__ import annotations
from typing import List, Union

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.atleast_2d)
def atleast_2d(*arys: PolyLike) -> Union[ndpoly, List[ndpoly]]:
    """
    View inputs as arrays with at least two dimensions.

    Args:
        arys:
            One or more array-like sequences. Non-array inputs are converted
            to arrays. Arrays that already have two or more dimensions are
            preserved.

    Return:
        An array, or list of arrays, each with ``a.ndim >= 2``. Copies are
        avoided where possible, and views with two or more dimensions are
        returned.

    Example:
        >>> numpoly.atleast_2d(numpoly.variable())
        polynomial([[q0]])
        >>> numpoly.atleast_2d(1, [2], [[3]])
        [polynomial([[1]]), polynomial([[2]]), polynomial([[3]])]

    """
    if len(arys) == 1:
        poly = numpoly.aspolynomial(arys[0])
        array = numpy.atleast_2d(poly.values)
        return numpoly.aspolynomial(array, names=poly.indeterminants)
    return [atleast_2d(ary) for ary in arys]



"""Find the lead exponents for each polynomial."""

from __future__ import annotations

import numpy
import numpoly

from ..baseclass import PolyLike


def lead_exponent(
    poly: PolyLike,
    graded: bool = False,
    reverse: bool = False,
) -> numpy.ndarray:
    """
    Find the lead exponents for each polynomial.

    As polynomials are not inherently sortable, values are sorted using the
    highest `lexicographical` ordering. Between the values that have the same
    highest ordering, the elements are sorted using the coefficients.

    Args:
        poly:
            Polynomial to locate exponents on.
        graded:
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``q0**2*q1**2*q2**2`` has an exponent sum of 6, and will
            therefore be consider larger than both ``q0**3*q1*q2``,
            ``q0*q1**3*q2`` and ``q0*q1*z**3``.
        reverse:
            Reverses lexicographical sorting meaning that ``q0*q1**3`` is
            considered bigger than ``q0**3*q1``, instead of the opposite.

    Return:
        Integer array with the largest exponents in the polynomials. The
        shape is ``poly.shape + (len(poly.names),)``. The extra dimension
        is used to indicate the exponent for the different indeterminants.

    Example:
        >>> q0 = numpoly.variable()
        >>> numpoly.lead_exponent([1, q0+1, q0**2+q0+1]).T
        array([[0, 1, 2]])
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.lead_exponent(
        ...     [1, q0, q1, q0*q1, q0**3-1]).T
        array([[0, 1, 0, 1, 3],
               [0, 0, 1, 1, 0]])

    """
    poly_ = numpoly.aspolynomial(poly)
    shape = poly_.shape
    poly = poly_.ravel()
    out = numpy.zeros(poly_.shape + (len(poly_.names),), dtype=int)
    if not poly_.size:
        return out
    for idx in numpoly.glexsort(poly_.exponents.T, graded=graded, reverse=reverse):
        out[poly_.coefficients[idx] != 0] = poly_.exponents[idx]
    return out.reshape(shape + (len(poly_.names),))



"""Load data from a text file."""

from __future__ import annotations
from typing import Callable, Dict, Optional, Sequence, Union
import re
from os import PathLike

import numpy
import numpy.typing
from numpy.lib.recfunctions import unstructured_to_structured
import numpoly

from .savetxt import HEADER_TEMPLATE
from ..baseclass import ndpoly

HEADER_REGEX = re.compile(
    HEADER_TEMPLATE.format(
        version=r"\S+", names=r"(\S+)", keys=r"(\S+)", shape=r"(\S*)"
    )
)


def loadtxt(
    fname: PathLike,
    dtype: numpy.typing.DTypeLike = float,
    comments: str = "# ",
    delimiter: Optional[str] = None,
    converters: Optional[Dict[int, Callable]] = None,
    skiprows: int = 0,
    usecols: Union[None, int, Sequence[int]] = None,
    unpack: bool = False,
    ndmin: int = 0,
    encoding: str = "bytes",
    max_rows: Optional[int] = None,
) -> ndpoly:
    """
    Load data from a text file.

    Each row in the text file must have the same number of values.

    Args:
        fname:
            File, filename, or generator to read.  If the filename extension
            is ``.gz`` or ``.bz2``, the file is first decompressed. Note that
            generators should return byte strings.
        dtype:
            Data-type of the resulting array; default: float.  If this is a
            structured data-type, the resulting array will be 1-dimensional,
            and each row will be interpreted as an element of the array.  In
            this case, the number of columns used must match the number of
            fields in the data-type.
        comments:
            The characters or list of characters used to indicate the start of
            a comment. None implies no comments. For backwards compatibility,
            byte strings will be decoded as 'latin1'. The default is '#'.
        delimiter:
            The string used to separate values. For backwards compatibility,
            byte strings will be decoded as 'latin1'. The default is
            whitespace.
        converters:
            A dictionary mapping column number to a function that will parse
            the column string into the desired value.  E.g., if column 0 is a
            date string: ``converters = {0: datestr2num}``.  Converters can
            also be used to provide a default value for missing data (but see
            also `genfromtxt`): ``converters = {3: lambda s: float(s.strip()
            or 0)}``.
        skiprows:
            Skip the first `skiprows` lines, including comments.
        usecols:
            Which columns to read, with 0 being the first. For example,
            ``usecols = (1,4,5)`` will extract the 2nd, 5th and 6th columns.
            The default, None, results in all columns being read.
        unpack:
            If True, the returned array is transposed, so that arguments may
            be unpacked using ``x, y, z = loadtxt(...)``.  When used with a
            structured data-type, arrays are returned for each field.
        ndmin:
            The returned array will have at least `ndmin` dimensions.
            Otherwise mono-dimensional axes will be squeezed. Legal values:
            0, 1 or 2.
        encoding:
            Encoding used to decode the inputfile. Does not apply to input
            streams. The special value 'bytes' enables backward compatibility
            workarounds that ensures you receive byte arrays as results if
            possible and passes 'latin1' encoded strings to converters.
            Override this value to receive unicode arrays and pass strings as
            input to converters.  If set to None the system default is used.
            The default value is 'bytes'.
        max_rows (Optional[int]): int, optional
            Read `max_rows` lines of content after `skiprows` lines. The
            default is to read all the lines.

    Return:
        Data read from the text file.

    Example:
        >>> q0, q1, q2 = numpoly.variable(3)
        >>> numpoly.savetxt("/tmp/poly.txt", q1)
        >>> numpoly.loadtxt("/tmp/poly.txt")
        polynomial(q1)
        >>> poly = numpoly.polynomial([[1, q0], [q0, q2**2 - 1]])
        >>> numpoly.savetxt("/tmp/poly.txt", poly)
        >>> numpoly.loadtxt("/tmp/poly.txt")
        polynomial([[1.0, q0],
                    [q0, q2**2-1.0]])

    """
    if isinstance(fname, (str, bytes, PathLike)):
        with open(fname) as src:
            header = src.readline()
    else:
        header = fname.readline()
    if isinstance(header, bytes):
        header = header.decode("utf-8")

    array = numpy.loadtxt(
        fname,
        dtype=dtype,
        comments=comments,
        delimiter=delimiter,
        converters=converters,
        skiprows=skiprows,
        usecols=usecols,
        unpack=unpack,
        ndmin=ndmin,
        max_rows=max_rows,
        encoding=encoding,
    )

    if header.startswith(comments + "numpoly:"):
        match = re.search(HEADER_REGEX, header)
        assert match is not None
        groups = match.groups()
        names = tuple(groups[0].split(","))
        keys = groups[1].split(",")
        shape = [int(idx) for idx in groups[2].split(",") if groups[2]]
        if not shape:
            array = array.reshape(-1)
        dtype = numpy.dtype([(key, array.dtype) for key in keys])
        struct = unstructured_to_structured(array, dtype)
        array = numpoly.polynomial(struct, names=names)
        array = numpoly.reshape(array, shape)

    return array



"""Functions for constructing polynomials."""

from .polynomial import polynomial
from .aspolynomial import aspolynomial
from .clean import (
    clean_attributes,
    remove_redundant_coefficients,
    remove_redundant_names,
)
from .from_attributes import polynomial_from_attributes
from .from_roots import polynomial_from_roots
from .monomial import monomial
from .symbols import symbols
from .variable import variable



"""Convert numpoly.ndpoly to sympy polynomial."""

from __future__ import annotations
from typing import Any

import numpy
import numpoly

from .baseclass import PolyLike


def to_sympy(poly: PolyLike) -> Any:
    """
    Convert numpoly object to sympy object, or array of sympy objects.

    Args:
        poly:
            Polynomial object to convert to sympy.

    Return:
        If scalar, a sympy expression object, or if array, numpy.array with
        expression object values.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([[1, q0**3], [q1-1, -3*q0]])
        >>> sympy_poly = to_sympy(poly)
        >>> sympy_poly
        array([[1, q0**3],
               [q1 - 1, -3*q0]], dtype=object)
        >>> type(sympy_poly.item(-1))
        <class 'sympy.core.mul.Mul'>

    """
    poly = numpoly.aspolynomial(poly)
    if poly.shape:
        return numpy.array([to_sympy(poly_) for poly_ in poly])
    from sympy import symbols  # type: ignore

    locals_ = dict(zip(poly.names, symbols(poly.names)))
    polynomial = eval(str(poly), locals_, {})  # pylint: disable=eval-used
    return polynomial



"""Return true division of the inputs, element-wise."""

from __future__ import division
from typing import Any, Optional

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements_ufunc

DIVIDE_ERROR_MSG = """
Divisor in division is a polynomial.
Polynomial division differs from numerical division;
Use ``numpoly.poly_divide`` to get polynomial division."""


@implements_ufunc(numpy.true_divide)
def true_divide(
    x1: PolyLike,
    x2: PolyLike,
    out: Optional[ndpoly] = None,
    where: numpy.typing.ArrayLike = True,
    **kwargs: Any,
) -> ndpoly:
    """
    Return true division of the inputs, element-wise.

    Instead of the Python traditional 'floor division', this returns a true
    division.  True division adjusts the output type to present the best
    answer, regardless of input types.

    Args:
        x1:
            Dividend array.
        x2:
            Divisor array. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Return:
        This is a scalar if both `x1` and `x2` are scalars.

    Raise:
        numpoly.baseclass.FeatureNotSupported:
            If `x2` contains indeterminants, numerical division is no longer
            possible and an error is raised instead. For polynomial
            division see ``numpoly.poly_divide``.

    Example:
        >>> q0q1q2 = numpoly.variable(3)
        >>> numpoly.true_divide(q0q1q2, 4)
        polynomial([0.25*q0, 0.25*q1, 0.25*q2])
        >>> numpoly.true_divide(q0q1q2, [1, 2, 4])
        polynomial([q0, 0.5*q1, 0.25*q2])

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    if not x2.isconstant():
        raise numpoly.FeatureNotSupported(DIVIDE_ERROR_MSG)
    x2 = x2.tonumpy()
    if out is None:
        out_ = numpoly.ndpoly(
            exponents=x1.exponents,
            shape=x1.shape,
            names=x1.indeterminants,
            dtype=numpy.common_type(x1, numpy.array(1.0)),
        )
    else:
        assert len(out) == 1
        out_ = out[0]
    assert isinstance(out_, numpoly.ndpoly)
    for key in x1.keys:
        out_[key] = 0
        numpy.true_divide(
            x1.values[key],
            x2,
            out=out_.values[key],
            where=numpy.asarray(where),
            **kwargs,
        )
    if out is None:
        out_ = numpoly.clean_attributes(out_)
    return out_



"""Return the indices of the minimum values along an axis."""

from __future__ import annotations
from typing import Any, Optional

import numpy
import numpoly

from ..baseclass import PolyLike
from ..dispatch import implements


@implements(numpy.argmin)
def argmin(
    a: PolyLike,
    axis: Optional[int] = None,
    out: Optional[numpy.ndarray] = None,
) -> Any:
    """
    Return the indices of the minimum values along an axis.

    As polynomials are not inherently sortable, values are sorted using the
    highest `lexicographical` ordering. Between the values that have the same
    highest ordering, the elements are sorted using the coefficients. This also
    ensures that the method behaves as expected with ``numpy.ndarray``.

    Args:
        a:
            Input array.
        axis:
            By default, the index is into the flattened array, otherwise along
            the specified axis.
        out:
            If provided, the result will be inserted into this array. It should
            be of the appropriate shape and dtype.

    Return:
        Array of indices into the array. It has the same shape as `a.shape`
        with the dimension along `axis` removed.

    Note:
        In case of multiple occurrences of the minimum values, the
        indices corresponding to the first occurrence are returned.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.argmin([13, 7])
        np.int64(1)
        >>> numpoly.argmin([1, q0, q0**2, q1])
        np.int64(0)
        >>> numpoly.argmin([q0*q1, q0, q1])
        np.int64(1)
        >>> numpoly.argmin([[3*q0**2, q0**2], [2*q0**2, 4*q0**2]], axis=0)
        array([1, 0])

    """
    a = numpoly.aspolynomial(a)
    options = numpoly.get_options()
    proxy = numpoly.sortable_proxy(
        a, graded=options["sort_graded"], reverse=options["sort_reverse"]
    )
    return numpy.argmin(proxy, axis=axis, out=out)



"""Copy values from one array to another, broadcasting as necessary."""

from __future__ import annotations
import logging

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.copyto)
def copyto(
    dst: ndpoly,
    src: PolyLike,
    casting: str = "same_kind",
    where: numpy.typing.ArrayLike = True,
) -> None:
    """
    Copy values from one array to another, broadcasting as necessary.

    Raises a TypeError if the `casting` rule is violated, and if
    `where` is provided, it selects which elements to copy.

    Args:
        dst:
            The array into which values are copied.
        src:
            The array from which values are copied.
        casting:
            Controls what kind of data casting may occur when copying.

            * 'no' means the data types should not be cast at all.
            * 'equiv' means only byte-order changes are allowed.
            * 'safe' means only casts which can preserve values are allowed.
            * 'same_kind' means only safe casts or casts within a kind,
                like float64 to float32, are allowed.
            * 'unsafe' means any data conversions may be done.
        where:
            A boolean array which is broadcasted to match the dimensions
            of `dst`, and selects elements to copy from `src` to `dst`
            wherever it contains the value True.

    Example:
        >>> q0 = numpoly.variable()
        >>> poly1 = numpoly.polynomial([1, q0**2, q0])
        >>> poly2 = numpoly.polynomial([2, q0, 3])
        >>> numpoly.copyto(poly1, poly2, where=[True, False, True])
        >>> poly1
        polynomial([2, q0**2, 3])
        >>> numpoly.copyto(poly1, poly2)
        >>> poly1
        polynomial([2, q0, 3])

    """
    logger = logging.getLogger(__name__)
    src = numpoly.aspolynomial(src)
    assert isinstance(dst, numpy.ndarray)
    if not isinstance(dst, numpoly.ndpoly):
        if dst.dtype.names is None:
            if src.isconstant():
                return numpy.copyto(dst, src.tonumpy(), casting=casting, where=where)
            raise ValueError(f"Could not convert src {src} to dst {dst}")
        if casting != "unsafe":
            raise ValueError(f"could not safely convert src {src} to dst {dst}")
        logger.warning("Copying ndpoly input into ndarray")
        logger.warning("You might need to cast `numpoly.polynomial(dst)`.")
        logger.warning("Indeterminant names might be lost.")
        dst_keys = dst.dtype.names
        dst_ = dst
    else:
        dst_keys = dst.keys
        src, _ = numpoly.align_indeterminants(src, dst)
        dst_ = dst.values

    missing_keys = set(src.keys).difference(dst_keys)
    if missing_keys:
        raise ValueError(f"memory layouts are incompatible: {missing_keys}")

    for key in dst_keys:
        if key in src.keys:
            numpy.copyto(dst_[key], src.values[key], casting=casting, where=where)
        else:
            numpy.copyto(
                dst_[key],
                numpy.array(False, dtype=dst_[key].dtype),
                casting=casting,
                where=where,
            )



"""Extract a diagonal or construct a diagonal array."""

from __future__ import annotations
import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.diag)
def diag(y: PolyLike, k: int = 0) -> ndpoly:
    """
    Extract a diagonal or construct a diagonal array.

    Args:
        v:
            If `v` is a 2-D array, return a copy of its `k`-th diagonal.
            If `v` is a 1-D array, return a 2-D array with `v` on the `k`-th
            diagonal.
        k:
            Diagonal in question. Use `k > 0` for diagonals above the main
            diagonal, and `k < 0` for diagonals below the main diagonal.

    Return:
        The extracted diagonal or constructed diagonal array.

    Example:
        >>> poly = numpoly.monomial(9).reshape(3, 3)
        >>> poly
        polynomial([[1, q0, q0**2],
                    [q0**3, q0**4, q0**5],
                    [q0**6, q0**7, q0**8]])
        >>> numpoly.diag(poly)
        polynomial([1, q0**4, q0**8])
        >>> numpoly.diag(poly, k=1)
        polynomial([q0, q0**5])
        >>> numpoly.diag(poly, k=-1)
        polynomial([q0**3, q0**7])

    """
    y = numpoly.aspolynomial(y)
    out = numpy.diag(y.values, k=k)
    return numpoly.polynomial(out, names=y.names)



"""Polynomials differentiation functions."""

from __future__ import annotations
from typing import Union

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike


def derivative(poly: PolyLike, *diffvars: Union[ndpoly, str, int]) -> ndpoly:
    """
    Polynomial differential operator.

    Args:
        poly:
            Polynomial to differentiate.
        diffvars:
            Singleton variables to take derivative off.

    Return:
        Same as ``poly`` but differentiated with respect to ``diffvars``.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([1, q0, q0*q1**2+1])
        >>> poly
        polynomial([1, q0, q0*q1**2+1])
        >>> numpoly.derivative(poly, "q0")
        polynomial([0, 1, q1**2])
        >>> numpoly.derivative(poly, 0, 1)
        polynomial([0, 0, 2*q1])
        >>> numpoly.derivative(poly, q0, q0, q0)
        polynomial([0, 0, 0])

    """
    poly = poly_ref = numpoly.aspolynomial(poly)

    for diffvar in diffvars:
        if isinstance(diffvar, str):
            idx = poly.names.index(diffvar)
        elif isinstance(diffvar, int):
            idx = diffvar
        else:
            diffvar = numpoly.aspolynomial(diffvar)
            exponents, names = numpoly.remove_redundant_names(
                diffvar.exponents, diffvar.names
            )
            assert names is not None and len(names) == 1, "one at the time"
            assert numpy.all(exponents == 1), "derivative variable assumes singletons"
            idx = poly.names.index(names[0])

        exponents = poly.exponents
        coefficients = [
            (exponent[idx] * coefficient.T).T
            for exponent, coefficient in zip(exponents, poly.coefficients)
        ]
        exponents[:, idx] -= 1
        assert not numpy.any(exponents < 0)

        poly = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            names=poly_ref.names,
        )
        poly, poly_ref = numpoly.align_polynomials(poly, poly_ref)
    return poly


def gradient(poly: PolyLike) -> ndpoly:
    """
    Polynomial gradient operator.

    Args:
        poly:
            Polynomial to differentiate. If polynomial vector is provided,
            the Jacobi-matrix is returned instead.

    Return:
        Same as ``poly`` but with an extra first dimensions, one for each
        variable in ``poly.indeterminants``, filled with gradient values.

    Example:
        >>> q0 = numpoly.variable()
        >>> poly = 5*q0**5+4
        >>> numpoly.gradient(poly)
        polynomial([25*q0**4])
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = 4*q0**3+2*q1**2+3
        >>> numpoly.gradient(poly)
        polynomial([12*q0**2, 4*q1])
        >>> poly = numpoly.polynomial([1, q0**3, q0*q1**2+1])
        >>> numpoly.gradient(poly)
        polynomial([[0, 3*q0**2, q1**2],
                    [0, 0, 2*q0*q1]])

    """
    poly = numpoly.aspolynomial(poly)
    polys = [derivative(poly, diffvar)[numpy.newaxis] for diffvar in poly.names]
    return numpoly.concatenate(polys, axis=0)


def hessian(poly: PolyLike) -> ndpoly:
    """
    Construct Hessian matrix of polynomials.

    Make Hessian matrix out of a polynomial. Tensor is returned if polynomial
    is a vector.

    Args:
        poly:
            Polynomial to differentiate.

    Return:
        Same as ``poly`` but with two extra dimensions, one for each
        variable in ``poly.indeterminants``, filled with Hessian values.

    Example:
        >>> q0 = numpoly.variable()
        >>> poly = 5*q0**5+4
        >>> numpoly.hessian(poly)
        polynomial([[100*q0**3]])
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = 4*q0**3+2*q1**2+3
        >>> numpoly.hessian(poly)
        polynomial([[24*q0, 0],
                    [0, 4]])
        >>> poly = numpoly.polynomial([1, q0, q0*q1**2+1])
        >>> numpoly.hessian(poly)
        polynomial([[[0, 0, 0],
                     [0, 0, 2*q1]],
        <BLANKLINE>
                    [[0, 0, 2*q1],
                     [0, 0, 2*q0]]])

    """
    return gradient(gradient(poly))



"""Clean up polynomial attributes."""

from __future__ import annotations
from typing import List, Optional, Sequence, Tuple, Union

import numpy
import numpy.typing

import numpoly
from ..baseclass import ndpoly


class PolynomialConstructionError(ValueError):
    """Error related to construction of polynomial."""


def clean_attributes(
    poly: ndpoly,
    retain_coefficients: Optional[bool] = None,
    retain_names: Optional[bool] = None,
) -> ndpoly:
    """
    Clean up polynomial attributes that does not affect representation.

    Some operations results in polynomial with structures that are redundant.
    This includes extra unused indeterminants and coefficients consisting of
    only zeros.

    Args:
        poly:
            Polynomial to clean up.
        retain_coefficients:
            Do not remove redundant coefficients. If omitted use global
            defaults.
        retain_names:
            Do not remove redundant names. If omitted use global defaults.

    Return:
        Same as `poly`, but with attributes cleaned up.

    Example:
        >>> q0, _ = numpoly.align_polynomials(*numpoly.variable(2))
        >>> q0.indeterminants
        polynomial([q0, q1])
        >>> q0.exponents
        array([[0, 1],
               [1, 0]], dtype=uint32)
        >>> q0 = numpoly.clean_attributes(
        ...     q0, retain_coefficients=False, retain_names=False)
        >>> q0.indeterminants
        polynomial([q0])
        >>> q0.exponents
        array([[1]], dtype=uint32)

    """
    return numpoly.ndpoly.from_attributes(
        exponents=poly.exponents,
        coefficients=poly.coefficients,
        names=poly.names,
        dtype=poly.dtype,
        retain_coefficients=retain_coefficients,
        retain_names=retain_names,
    )


def postprocess_attributes(
    exponents: numpy.typing.ArrayLike,
    coefficients: Sequence[numpy.typing.ArrayLike],
    names: Union[None, str, Tuple[str, ...], ndpoly] = None,
    retain_coefficients: Optional[bool] = None,
    retain_names: Optional[bool] = None,
) -> Tuple[numpy.ndarray, List[numpy.ndarray], Optional[Tuple[str, ...]]]:
    """
    Clean up polynomial attributes.

    Args:
        exponents:
            The exponents in an integer array with shape ``(N, D)``, where
            ``N`` is the number of terms in the polynomial sum and ``D`` is
            the number of dimensions.
        coefficients:
            The polynomial coefficients. Must correspond to `exponents` by
            having the same length ``N``.
        names:
            The indeterminant names, either as string names or as
            simple polynomials. Must correspond to the exponents by having
            length ``D``.
        retain_coefficients:
            Do not remove redundant coefficients. If omitted use global
            defaults.
        retain_names:
            Do not remove redundant names. If omitted use global defaults.

    Return:
        Same as inputs `exponents`, `coefficients` and `names`, but
        post-processed.

    """
    exponents = numpy.asarray(exponents)
    if exponents.ndim != 2:
        raise PolynomialConstructionError(
            f"expected exponents.ndim == 2; found {exponents.ndim}"
        )

    coefficients_ = [numpy.asarray(coefficient) for coefficient in coefficients]
    if coefficients_ and len(exponents) != len(coefficients_):
        raise PolynomialConstructionError(
            "expected len(exponents) == len(coefficients_); "
            f"found {len(exponents)} != {len(coefficients_)}"
        )

    if retain_coefficients is None:
        retain_coefficients = numpoly.get_options()["retain_coefficients"]
    if not retain_coefficients and coefficients_:
        exponents, coefficients_ = remove_redundant_coefficients(
            exponents, coefficients_
        )

    if isinstance(names, numpoly.ndpoly):
        names = names.names
    if isinstance(names, str):
        if exponents.shape[1] > 1:
            names = tuple(f"{names}{idx}" for idx in range(exponents.shape[1]))
        else:
            names = (names,)
    if names:
        if len(names) != exponents.shape[1]:
            raise PolynomialConstructionError(
                "Name length incompatible exponent length; "
                f"len({names}) != {exponents.shape[1]}"
            )
        if sorted(set(names)) != sorted(names):
            raise PolynomialConstructionError(f"Duplicate indeterminant names: {names}")

    if retain_names is None:
        retain_names = numpoly.get_options()["retain_names"]
    if not retain_names:
        exponents, names = remove_redundant_names(exponents, names)

    exponents_, count = numpy.unique(exponents, return_counts=True, axis=0)
    if numpy.any(count > 1):
        raise PolynomialConstructionError(
            f"Duplicate exponent keys found: {exponents_[count > 1][0]}"
        )

    return numpy.asarray(exponents), list(coefficients_), names


def remove_redundant_coefficients(
    exponents: numpy.typing.ArrayLike,
    coefficients: Sequence[numpy.typing.ArrayLike],
) -> Tuple[numpy.ndarray, List[numpy.ndarray]]:
    """
    Remove coefficients if they are redundant to the polynomial representation.

    Will always keep at least one coefficient.

    Args:
        exponents:
            The exponents in an integer array with shape ``(N, D)``, where
            ``N`` is the number of terms in the polynomial sum and ``D`` is
            the number of dimensions.
        coefficients:
            The polynomial coefficients. Must correspond to `exponents` by
            having the same length ``N``.

    Return:
        Same as inputs, but with redundant exponent rows and coefficients
        removed. This corresponds to ``N`` being lowered.

    Example:
        >>> exponents, coefficients = remove_redundant_coefficients(
        ...     [[0, 0], [0, 1]], [1, 2])
        >>> exponents
        array([[0, 0],
               [0, 1]])
        >>> coefficients
        [array(1), array(2)]
        >>> remove_redundant_coefficients(
        ...     [[0, 0], [0, 1]], [1, 0])
        (array([[0, 0]]), [array(1)])
        >>> remove_redundant_coefficients([[0]], [[]])
        (array([[0]]), [array([], dtype=float64)])

    """
    exponents_ = numpy.asarray(exponents)
    coefficients_ = [numpy.asarray(coefficient) for coefficient in coefficients]
    assert len(exponents_) == len(coefficients_), (exponents, coefficients)

    elements = list(
        zip(
            *[
                (exponent, coefficient)
                for exponent, coefficient in zip(exponents_, coefficients_)
                if numpy.any(coefficient) or not numpy.any(exponent)
            ]
        )
    )
    if not elements:
        exponents_ = numpy.zeros((1, exponents_.shape[-1]), dtype="uint32")
        coefficients_ = [numpy.zeros_like(coefficients_[0])]

    else:
        exponents_ = numpy.asarray(elements[0], dtype=int)
        coefficients_ = list(elements[1])
    return exponents_, coefficients_


def remove_redundant_names(
    exponents: numpy.typing.ArrayLike,
    names: Optional[Sequence[str]],
) -> Tuple[numpy.ndarray, Optional[Tuple[str, ...]]]:
    """
    Remove names if they are redundant to the polynomial representation.

    Will always keep at least one dimension.

    Args:
        exponents:
            The exponents in an integer array with shape ``(N, D)``, where
            ``N`` is the number of terms in the polynomial sum and ``D`` is
            the number of dimensions.
        names:
            The indeterminant names, either as string names or as
            simple polynomials. Must correspond to the exponents by having
            length ``D``.

    Return:
        Same as inputs, but with redundant exponent columns and names removed.
        This corresponds to ``D`` being lowered.

    Example:
        >>> exponents, names = remove_redundant_names(
        ...     [[0, 0], [1, 1]], ["q0", "q1"])
        >>> exponents
        array([[0, 0],
               [1, 1]])
        >>> names
        ('q0', 'q1')
        >>> exponents, names = remove_redundant_names(
        ...     [[0, 0], [0, 1]], ["q0", "q1"])
        >>> exponents
        array([[0],
               [1]])
        >>> names
        ('q1',)
        >>> exponents, names = remove_redundant_names(
        ...     [[0, 0]], ["q0", "q1"])
        >>> exponents
        array([[0]])
        >>> names
        ('q0',)

    """
    exponents_ = numpy.asarray(exponents)
    indices = numpy.any(exponents_ != 0, 0)
    if not numpy.any(indices):
        indices[0] = True

    exponents_ = exponents_[:, indices]
    if names is None:
        return exponents_, names
    names_ = tuple(numpy.array(names)[indices].tolist())
    return exponents_, names_



"""Element-wise minimum of array elements."""

from __future__ import annotations
from typing import Any, Optional

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.minimum)
def minimum(
    x1: PolyLike,
    x2: PolyLike,
    out: Optional[ndpoly] = None,
    **kwargs: Any,
) -> ndpoly:
    """
    Element-wise minimum of array elements.

    Compare two arrays and returns a new array containing the element-wise
    minima. If one of the elements being compared is a NaN, then that
    element is returned. If both elements are NaNs then the first is
    returned. The latter distinction is important for complex NaNs, which
    are defined as at least one of the real or imaginary parts being a NaN.
    The net effect is that NaNs are propagated.

    Args:
        x1, x2 :
            The arrays holding the elements to be compared. If ``x1.shape !=
            x2.shape``, they must be broadcastable to a common shape (which
            becomes the shape of the output).
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Return:
        The minimum of `x1` and `x2`, element-wise. This is a scalar if
        both `x1` and `x2` are scalars.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.minimum(3, 4)
        polynomial(3)
        >>> numpoly.minimum(4*q0, 3*q0)
        polynomial(3*q0)
        >>> numpoly.minimum(q0, q1)
        polynomial(q0)
        >>> numpoly.minimum(q0**2, q0)
        polynomial(q0)
        >>> numpoly.minimum([1, q0, q0**2, q0**3], q1)
        polynomial([1, q0, q1, q1])
        >>> numpoly.minimum(q0+1, q0-1)
        polynomial(q0-1)

    """
    del out
    x1, x2 = numpoly.align_polynomials(x1, x2)
    coefficients1 = x1.coefficients
    coefficients2 = x2.coefficients
    out_ = numpy.zeros(x1.shape, dtype=bool)

    options = numpoly.get_options()
    for idx in numpoly.glexsort(
        x1.exponents.T, graded=options["sort_graded"], reverse=options["sort_reverse"]
    ):
        indices = (coefficients1[idx] != 0) | (coefficients2[idx] != 0)
        indices = coefficients1[idx] != coefficients2[idx]
        out_[indices] = (coefficients1[idx] < coefficients2[idx])[indices]
    return numpoly.where(out_, x1, x2)



"""Return the indices of the elements that are non-zero."""

from __future__ import annotations
from typing import Any, Tuple

import numpy
import numpoly

from ..baseclass import PolyLike
from ..dispatch import implements


@implements(numpy.nonzero)
def nonzero(x: PolyLike, **kwargs: Any) -> Tuple[numpy.ndarray, ...]:
    """
    Return the indices of the elements that are non-zero.

    Args:
        x:
            Input array.

    Return:
        Indices of elements that are non-zero.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([[3*q0, 0, 0],
        ...                            [0, 4*q1, 0],
        ...                            [5*q0+q1, 6*q0, 0]])
        >>> poly
        polynomial([[3*q0, 0, 0],
                    [0, 4*q1, 0],
                    [q1+5*q0, 6*q0, 0]])
        >>> numpoly.nonzero(poly)
        (array([0, 1, 2, 2]), array([0, 1, 0, 1]))
        >>> poly[numpoly.nonzero(poly)]
        polynomial([3*q0, 4*q1, q1+5*q0, 6*q0])

    """
    x = numpoly.aspolynomial(x)
    return numpy.nonzero(numpy.any(numpy.asarray(x.coefficients), axis=0))



"""Matrix product of two arrays."""

from __future__ import annotations
from typing import Any, Optional

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements

ERROR_MESSAGE = """\
matmul: Input operand %d does not have enough dimensions \
(has 0, gufunc core with signature (n?,k),(k,m?)->(n?,m?) requires 1)
"""


@implements(numpy.matmul)
def matmul(
    x1: PolyLike,
    x2: PolyLike,
    out: Optional[ndpoly] = None,
    **kwargs: Any,
) -> ndpoly:
    """
    Matrix product of two arrays.

    Args:
        x1, x2:
            Input arrays, scalars not allowed.
        out:
            A location into which the result is stored. If provided, it must
            have a shape that matches the signature `(n,k),(k,m)->(n,m)`.
            If not provided or `None`, a freshly-allocated array is returned.

    Return:
        The matrix product of the inputs. This is a scalar only when both
        x1, x2 are 1-d vectors.

    Raise:
        ValueError:
            If the last dimension of `x1` is not the same size as
            the second-to-last dimension of `x2`.

    Example:
        >>> poly = numpoly.variable(4).reshape(2, 2)
        >>> poly
        polynomial([[q0, q1],
                    [q2, q3]])
        >>> numpoly.matmul(poly, [[0, 1], [2, 3]])
        polynomial([[2*q1, 3*q1+q0],
                    [2*q3, 3*q3+q2]])
        >>> numpoly.matmul(poly, [4, 5])
        polynomial([[4*q1+4*q0, 5*q1+5*q0],
                    [4*q3+4*q2, 5*q3+5*q2]])
        >>> numpoly.matmul(*poly)
        polynomial([q1*q2+q0*q2, q1*q3+q0*q3])

    """
    x1 = numpoly.aspolynomial(x1)
    x2 = numpoly.aspolynomial(x2)
    if not x1.shape:
        raise ValueError(ERROR_MESSAGE % 0)
    if not x2.shape:
        raise ValueError(ERROR_MESSAGE % 1)
    x1 = numpoly.reshape(x1, x1.shape + (1,))
    x2 = numpoly.reshape(x2, x2.shape[:-2] + (1,) + x2.shape[-2:])
    x1, x2 = numpoly.broadcast_arrays(x1, x2)
    out_ = numpoly.multiply(x1, x2, out=out, **kwargs)
    return numpoly.sum(out_, axis=-2)



"""Return element-wise quotient and remainder simultaneously."""

from __future__ import annotations, division
from typing import Any, Optional, Tuple

import numpy
import numpy.typing

import numpoly

from ...baseclass import ndpoly, PolyLike
from ...dispatch import implements_function


@implements_function(numpy.divmod)
def poly_divmod(
    dividend: PolyLike,
    divisor: PolyLike,
    out: Tuple[Optional[ndpoly], Optional[ndpoly]] = (None, None),
    where: numpy.typing.ArrayLike = True,
    **kwargs: Any,
) -> Tuple[ndpoly, ndpoly]:
    """
    Return element-wise quotient and remainder simultaneously.

    ``numpoly.divmod(x, y)`` is equivalent to ``(x / y, x % y)``, but faster
    because it avoids redundant work. It is used to implement the Python
    built-in function ``divmod`` on Numpoly arrays.

    Note:
        Unlike numbers, this returns the polynomial division and polynomial
        remainder. This means that this function is _not_ backwards compatible
        with ``numpy.divmod`` for constants. For example:
        ``numpy.divmod(11, 2) == (5, 1)`` while
        ``numpoly.divmod(11, 2) == (5.5, 0)``.

    Args:
        dividend:
            The array being divided.
        divisor:
            Array that that will divide the dividend.
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Return:
        Element-wise quotient and remainder resulting from floor division. This
        is a scalar if both `x1` and `x2` are scalars.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> denominator = [q0*q1**2+2*q0**3*q1**2, -2+q0*q1**2]
        >>> numerator = -2+q0*q1**2
        >>> floor, remainder = numpoly.poly_divmod(
        ...     denominator, numerator)
        >>> floor
        polynomial([2.0*q0**2+1.0, 1.0])
        >>> remainder
        polynomial([4.0*q0**2+2.0, 0.0])
        >>> floor*numerator+remainder
        polynomial([2.0*q0**3*q1**2+q0*q1**2, q0*q1**2-2.0])

    """
    assert where is True, "changing 'where' is not supported."
    dividend_, divisor = numpoly.align_polynomials(dividend, divisor)

    if not dividend_.shape:
        floor, remainder = poly_divmod(
            dividend_.ravel(),
            divisor.ravel(),
            out=out,
            where=where,
            **kwargs,
        )
        return floor[0], remainder[0]

    quotient = numpoly.zeros(dividend_.shape)
    while True:
        candidates = get_division_candidate(dividend_, divisor)
        if candidates is None:
            break
        idx1, idx2, include, candidate = candidates

        exponent_diff = dividend_.exponents[idx1] - divisor.exponents[idx2]
        candidate = candidate * numpoly.prod(divisor.indeterminants**exponent_diff, 0)
        key = dividend_.keys[idx1]

        quotient = numpoly.add(quotient, numpoly.where(include, candidate, 0), **kwargs)
        dividend_ = numpoly.subtract(
            dividend_, numpoly.where(include, divisor * candidate, 0), **kwargs
        )

        # ensure the candidate values are actual zero
        if key in dividend_.keys:
            dividend_.values[key][include] = 0

        dividend_, divisor = numpoly.align_polynomials(dividend_, divisor)

    return quotient, dividend_


def get_division_candidate(
    x1: ndpoly,
    x2: ndpoly,
    cutoff: float = 1e-30,
) -> Optional[Tuple[int, int, numpy.ndarray, numpy.ndarray]]:
    """
    Find the next exponent candidate pair in the iterative subtraction process.

    Args:
        x1:
            The array being divided.
        x2:
            Array that that will divide the dividend.
        cutoff:
            Threshold for when a value is so small that it is no longer a valid
            candidate. Required to avoid infinity loops in some edge cases.

    Return:
        Indices to the exponent candidate in respectively `x1` and `x2`
        that should be used in the divide process, and an boolean array to
        indicate for which coefficients these candidates are valid.

    """
    # Look for exponent candidates among divisors
    for idx2 in reversed(numpy.lexsort(x2.exponents.T)):
        exponent2 = x2.exponents[idx2]

        # Include coefficients where idx2 is non-zero and any potential
        # candidates that is a better fit has coefficient zero. Exponent needs
        # to be the biggest one around.
        include2 = numpy.ones(x2.shape, dtype=bool)
        for idx, exponent in enumerate(x2.exponents):
            if numpy.all(exponent2 <= exponent):
                include2 &= (x2.coefficients[idx] == 0) ^ (idx == idx2)
        if not numpy.any(include2):
            continue

        # Look for exponent candidates among dividend
        for idx1 in reversed(numpy.lexsort(x1.exponents.T)):
            exponent1 = x1.exponents[idx1]

            # Skip situations where numerator has larger exponent than the
            # denominator.
            if numpy.any(exponent1 < exponent2):
                continue

            # Skip if all relevant coefficients of interest are zero.
            include1 = x1.coefficients[idx1] != 0
            include = include1 & include2
            if not numpy.any(include):
                continue

            # construct candidate
            candidate = x1.coefficients[idx1] / numpy.where(
                include, x2.coefficients[idx2], 1
            )

            # really big relative error makes division algorithm
            # into a convergence strategy which needs a cutoff.
            if numpy.all(numpy.abs(candidate) < cutoff):
                continue

            return idx1, idx2, include, candidate

    # No valid candidate pair found; Assume we are done.
    return None



# cvalues.pyx

# Import Cython and types from standard Python
import numpy as np
cimport numpy as np
from libc.stdint cimport int64_t, uint8_t, uint32_t


cdef void cset_bool_values_1d(
        uint8_t [::1] coeffs, 
        str name,
        np.ndarray out,
):
  
    cdef Py_ssize_t i, byte_row, nrow, value_offset

    # Declare pointers to the structured array 'out'
    cdef char *base_ptr = <char *> np.PyArray_DATA(out) 
    cdef char *data_ptr
    cdef uint8_t *value_ptr
    
    # Size for rows (in Bytes and Int)
    value_offset = <Py_ssize_t> out.dtype.fields[name][1]
    byte_row = out.strides[0]
    nrow = out.size
    for i in range(nrow):
        data_ptr = base_ptr + i * byte_row
        value_ptr = <uint8_t *> (data_ptr + value_offset)
        value_ptr[0] = coeffs[i]


cdef void cset_uint32_values_1d(
        uint32_t [::1] coeffs, 
        str name,
        np.ndarray out,
):
  
    cdef Py_ssize_t i, byte_row, nrow, value_offset

    # Declare pointers to the structured array 'out'
    cdef char *base_ptr = <char *> np.PyArray_DATA(out) 
    cdef char *data_ptr
    cdef uint32_t *value_ptr
    
    # Size for rows (in Bytes and Int)
    value_offset = <Py_ssize_t> out.dtype.fields[name][1]
    byte_row = out.strides[0]
    nrow = out.size
    for i in range(nrow):
        data_ptr = base_ptr + i * byte_row
        value_ptr = <uint32_t *> (data_ptr + value_offset)
        value_ptr[0] = coeffs[i]


cdef void cset_int_values_1d(
        int64_t [::1] coeffs, 
        str name,
        np.ndarray out,
):
  
    cdef Py_ssize_t i, byte_row, nrow, value_offset

    # Declare pointers to the structured array 'out'
    cdef char *base_ptr = <char *> np.PyArray_DATA(out) 
    cdef char *data_ptr
    cdef int64_t *value_ptr
    
    # Size for rows (in Bytes and Int)
    value_offset = <Py_ssize_t> out.dtype.fields[name][1]
    byte_row = out.strides[0]
    nrow = out.size
    for i in range(nrow):
        data_ptr = base_ptr + i * byte_row
        value_ptr = <int64_t *> (data_ptr + value_offset)
        value_ptr[0] = coeffs[i]


cdef void cset_float_values_1d(
        double [::1] coeffs, 
        str name,
        np.ndarray out,
):
  
    cdef Py_ssize_t i, byte_row, nrow, value_offset

    # Declare pointers to the structured array 'out'
    cdef char *base_ptr = <char *> np.PyArray_DATA(out) 
    cdef char *data_ptr
    cdef double *value_ptr
    
    # Size for rows (in Bytes and Int)
    value_offset = <Py_ssize_t> out.dtype.fields[name][1]
    byte_row = out.strides[0]
    nrow = out.size
    for i in range(nrow):
        data_ptr = base_ptr + i * byte_row
        value_ptr = <double *> (data_ptr + value_offset)
        value_ptr[0] = coeffs[i]


cdef void cset_complex_values_1d(
        complex [::1] coeffs, 
        str name,
        np.ndarray out,
):
  
    cdef Py_ssize_t i, byte_row, nrow, value_offset

    # Declare pointers to the structured array 'out'
    cdef char *base_ptr = <char *> np.PyArray_DATA(out) 
    cdef char *data_ptr
    cdef complex *value_ptr
    
    # Size for rows (in Bytes and Int)
    value_offset = <Py_ssize_t> out.dtype.fields[name][1]
    byte_row = out.strides[0]
    nrow = out.size
    for i in range(nrow):
        data_ptr = base_ptr + i * byte_row
        value_ptr = <complex *> (data_ptr + value_offset)
        value_ptr[0] = coeffs[i]


cpdef cset_values(
        coeffs,
        str name,
        np.ndarray out,
):
    if coeffs.dtype == np.bool_:
        cset_bool_values_1d(coeffs, name, out)
    elif coeffs.dtype == np.uint32:
        cset_uint32_values_1d(coeffs, name, out)
    elif coeffs.dtype == np.int64:
        cset_int_values_1d(coeffs, name, out)
    elif coeffs.dtype == np.float64:
        cset_float_values_1d(coeffs, name, out)
    elif coeffs.dtype == np.complex128:
        cset_complex_values_1d(coeffs, name, out)
    else:
        ValueError("TYPE not implemented")


cdef void cadd_bool_values_1d(
        uint8_t [::1] coeffs, 
        str name,
        np.ndarray out,
):
  
    cdef Py_ssize_t i, byte_row, nrow, value_offset

    # Declare pointers to the structured array 'out'
    cdef char *base_ptr = <char *> np.PyArray_DATA(out) 
    cdef char *data_ptr
    cdef uint8_t *value_ptr
    
    # Size for rows (in Bytes and Int)
    value_offset = <Py_ssize_t> out.dtype.fields[name][1]
    byte_row = out.strides[0]
    nrow = out.size
    for i in range(nrow):
        data_ptr = base_ptr + i * byte_row
        value_ptr = <uint8_t *> (data_ptr + value_offset)
        value_ptr[0] += coeffs[i]


cdef void cadd_uint32_values_1d(
        uint32_t [::1] coeffs, 
        str name,
        np.ndarray out,
):
  
    cdef Py_ssize_t i, byte_row, nrow, value_offset

    # Declare pointers to the structured array 'out'
    cdef char *base_ptr = <char *> np.PyArray_DATA(out) 
    cdef char *data_ptr
    cdef uint32_t *value_ptr
    
    # Size for rows (in Bytes and Int)
    value_offset = <Py_ssize_t> out.dtype.fields[name][1]
    byte_row = out.strides[0]
    nrow = out.size
    for i in range(nrow):
        data_ptr = base_ptr + i * byte_row
        value_ptr = <uint32_t *> (data_ptr + value_offset)
        value_ptr[0] += coeffs[i]


cdef void cadd_int_values_1d(
        int64_t [::1] coeffs, 
        str name,
        np.ndarray out,
):
  
    cdef Py_ssize_t i, byte_row, nrow, value_offset

    # Declare pointers to the structured array 'out'
    cdef char *base_ptr = <char *> np.PyArray_DATA(out) 
    cdef char *data_ptr
    cdef int64_t *value_ptr
    
    # Size for rows (in Bytes and Int)
    value_offset = <Py_ssize_t> out.dtype.fields[name][1]
    byte_row = out.strides[0]
    nrow = out.size
    for i in range(nrow):
        data_ptr = base_ptr + i * byte_row
        value_ptr = <int64_t *> (data_ptr + value_offset)
        value_ptr[0] += coeffs[i]


cdef void cadd_float_values_1d(
        double [::1] coeffs, 
        str name,
        np.ndarray out,
):
  
    cdef Py_ssize_t i, byte_row, nrow, value_offset

    # Declare pointers to the structured array 'out'
    cdef char *base_ptr = <char *> np.PyArray_DATA(out) 
    cdef char *data_ptr
    cdef double *value_ptr
    
    # Size for rows (in Bytes and Int)
    value_offset = <Py_ssize_t> out.dtype.fields[name][1]
    byte_row = out.strides[0]
    nrow = out.size
    for i in range(nrow):
        data_ptr = base_ptr + i * byte_row
        value_ptr = <double *> (data_ptr + value_offset)
        value_ptr[0] += coeffs[i]


cdef void cadd_complex_values_1d(
        complex [::1] coeffs, 
        str name,
        np.ndarray out,
):
  
    cdef Py_ssize_t i, byte_row, nrow, value_offset

    # Declare pointers to the structured array 'out'
    cdef char *base_ptr = <char *> np.PyArray_DATA(out) 
    cdef char *data_ptr
    cdef complex *value_ptr
    
    # Size for rows (in Bytes and Int)
    value_offset = <Py_ssize_t> out.dtype.fields[name][1]
    byte_row = out.strides[0]
    nrow = out.size
    for i in range(nrow):
        data_ptr = base_ptr + i * byte_row
        value_ptr = <complex *> (data_ptr + value_offset)
        value_ptr[0] += coeffs[i]


cpdef cadd_values(
        coeffs,
        str name,
        np.ndarray out,
):
    if coeffs.dtype == np.bool_:
        cadd_bool_values_1d(coeffs, name, out)
    elif coeffs.dtype == np.uint32:
        cadd_uint32_values_1d(coeffs, name, out)
    elif coeffs.dtype == np.int64:
        cadd_int_values_1d(coeffs, name, out)
    elif coeffs.dtype == np.float64:
        cadd_float_values_1d(coeffs, name, out)
    elif coeffs.dtype == np.complex128:
        cadd_complex_values_1d(coeffs, name, out)
    else:
        ValueError("Type not implemented")



"""Attempt to cast an object into a polynomial array."""

from __future__ import annotations
from typing import Optional, Tuple, Union

import numpy
import numpy.typing

import numpoly

from .compose import compose_polynomial_array
from ..baseclass import ndpoly, PolyLike


def polynomial(
    poly_like: PolyLike = 0,
    names: Union[None, str, Tuple[str, ...], ndpoly] = None,
    dtype: Optional[numpy.typing.DTypeLike] = None,
    allocation: Optional[int] = None,
) -> ndpoly:
    """
    Attempt to cast an object into a polynomial array.

    Supports various casting options:

    ==================  =======================================================
    ``dict``            Keys are tuples that represent polynomial exponents,
                        and values are numpy arrays that represents polynomial
                        coefficients.
    ``numpoly.ndpoly``  Copy of the polynomial.
    ``numpy.ndarray``   Constant term polynomial.
    ``sympy.Poly``      Convert polynomial from ``sympy`` to ``numpoly``,
                        if possible.
    ``Iterable``        Multivariate array construction.
    structured array    Assumes that the input are raw polynomial core and can
                        be used to construct a polynomial without changing the
                        data. Used for developer convenience.
    ==================  =======================================================

    Args:
        poly_like:
            Input to be converted to a `numpoly.ndpoly` polynomial type.
        names:
            Name of the indeterminant variables. If possible to infer from
            ``poly_like``, this argument will be ignored.
        dtype:
            Data type used for the polynomial coefficients.
        allocation:
            The maximum number of polynomial exponents. If omitted, use
            length of exponents for allocation.

    Return:
        Polynomial based on input ``poly_like``.

    Example:
        >>> numpoly.polynomial({(1,): 1})
        polynomial(q0)
        >>> q0, q1 = numpoly.variable(2)
        >>> q0**2+q0*q1+2
        polynomial(q0*q1+q0**2+2)
        >>> -3*q0+q0**2+q1
        polynomial(q0**2+q1-3*q0)
        >>> numpoly.polynomial([q0*q1, q0, q1])
        polynomial([q0*q1, q0, q1])
        >>> numpoly.polynomial([1, 2, 3])
        polynomial([1, 2, 3])
        >>> import sympy
        >>> q0_, q1_ = sympy.symbols("q0, q1")
        >>> numpoly.polynomial(3*q0_*q1_-4+q0_**5)
        polynomial(q0**5+3*q0*q1-4)

    """
    if isinstance(poly_like, dict):
        poly = numpoly.ndpoly(exponents=[(0,)], shape=())
        exponents, coefficients = zip(*list(poly_like.items()))
        poly = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            names=names,
            dtype=dtype,
            allocation=allocation,
        )

    elif isinstance(poly_like, numpoly.ndpoly):
        if names is None:
            names = poly_like.names
        poly = numpoly.ndpoly.from_attributes(
            exponents=poly_like.exponents,
            coefficients=poly_like.coefficients,
            names=names,
            dtype=dtype,
            allocation=allocation,
        )

    # assume polynomial converted to structured array
    elif isinstance(poly_like, numpy.ndarray) and poly_like.dtype.names:
        keys = numpy.asarray(poly_like.dtype.names, dtype="U")
        keys = numpy.array(keys, dtype=f"U{numpy.max(numpy.char.str_len(keys))}")
        exponents = keys.view(numpy.uint32) - numpoly.ndpoly.KEY_OFFSET
        exponents = exponents.reshape(len(keys), -1)

        coefficients = [poly_like[key] for key in poly_like.dtype.names]
        poly = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            names=names,
            allocation=allocation,
        )

    elif isinstance(poly_like, (int, float, numpy.ndarray, numpy.generic)):
        poly = numpoly.ndpoly.from_attributes(
            exponents=[(0,)],
            coefficients=[numpy.asarray(poly_like)],
            names=names,
            dtype=dtype,
            allocation=allocation,
        )

    # handler for sympy objects
    elif hasattr(poly_like, "as_poly"):
        poly_like = poly_like.as_poly()  # type: ignore
        exponents = poly_like.monoms()  # type: ignore
        coefficients = [
            int(coeff) if coeff.is_integer else float(coeff)  # type: ignore
            for coeff in poly_like.coeffs()  # type: ignore
        ]
        names = [str(elem) for elem in poly_like.gens]  # type: ignore
        poly = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=coefficients,
            names=names,
            allocation=allocation,
        )

    else:
        poly = compose_polynomial_array(
            arrays=poly_like,  # type: ignore
            dtype=dtype,
            allocation=allocation,
        )

    return poly



"""Test for `numpoly.construct`."""
from pytest import raises
import numpy

import sympy
import numpoly
from numpoly.construct.clean import (
    postprocess_attributes, PolynomialConstructionError)

XY = numpoly.variable(2)
X, Y = numpoly.symbols("q0"), numpoly.symbols("q1")


def test_postprocess_attributes():
    """Test related to polynomial construction from attributes."""
    with raises(PolynomialConstructionError):  # exponent.ndim too small
        postprocess_attributes(numpy.arange(2), [1, 2])
    with raises(PolynomialConstructionError):  # exponent.ndim too large
        postprocess_attributes(numpy.arange(24).reshape(2, 3, 4), [1, 2])
    with raises(PolynomialConstructionError):  # exponents len incompatible with coefficients
        postprocess_attributes(numpy.arange(6).reshape(2, 3), [1, 2, 3])
    with raises(PolynomialConstructionError):  # duplicate exponents
        postprocess_attributes([[1], [1]], [1, 2])
    with raises(PolynomialConstructionError):  # exponents len incompatible with name length
        postprocess_attributes([[1], [2]], [1, 2], ["x", "y", "z"])
    with raises(PolynomialConstructionError):  # duplicate names
        postprocess_attributes([[1, 1]], [1], ["x", "x"])


def test_aspolynomial():
    poly = 2*X-Y+1
    assert poly == numpoly.aspolynomial(poly)
    assert poly == numpoly.aspolynomial(poly, names=XY)
    assert poly == numpoly.aspolynomial(poly.todict(), names=XY)
    assert poly == numpoly.aspolynomial(poly, names=("q0", "q1"))
    with numpoly.global_options(
            varname_filter=r"\w+", force_number_suffix=False):
        assert numpy.all(
            numpoly.symbols("Z:2") == numpoly.aspolynomial(XY, names="Z"))
    assert poly == numpoly.aspolynomial(poly.todict(), names=("q0", "q1"))
    assert poly != numpoly.aspolynomial(poly.todict(), names=("q1", "q0"))
    assert X == numpoly.aspolynomial(Y, names="q0")
    assert poly != numpoly.aspolynomial(poly.todict(), names="q0")
    assert isinstance(numpoly.aspolynomial([1, 2, 3]), numpoly.ndpoly)
    assert numpy.all(numpoly.aspolynomial([1, 2, 3]) == [1, 2, 3])


def test_monomial():
    assert not numpoly.monomial(0).size
    assert numpoly.monomial(1) == 1
    assert numpy.all(numpoly.monomial(2, dimensions="q0") == [1, X])
    assert numpoly.monomial(1, 2, dimensions="q0") == X
    assert numpoly.monomial(1, 2, dimensions=None) == X


def test_polynomial():
    assert numpoly.polynomial() == 0
    assert numpoly.polynomial({(0,): 4}) == 4
    assert numpoly.polynomial({(1,): 5}, names="q0") == 5*X
    assert numpoly.polynomial(
        {(0, 1): 2, (1, 0): 3}, names=("q0", "q1")) == 3*X+2*Y
    with numpoly.global_options(varname_filter=r"\w+"):
        assert numpy.all(numpoly.polynomial(
            {(0, 1): [0, 1], (1, 0): [1, 0]}, names="Q"
        ) == numpoly.symbols("Q0 Q1"))
    assert numpoly.polynomial(X) == X
    assert numpoly.polynomial(numpy.array((3,), dtype=[(";", int)])) == 3
    assert numpoly.polynomial(5.5) == 5.5
    assert numpoly.polynomial(sympy.symbols("q0")) == X
    assert numpy.all(numpoly.polynomial([1, 2, 3]) == [1, 2, 3])
    assert numpy.all(numpoly.polynomial([[1, 2], [3, 4]]) == [[1, 2], [3, 4]])
    assert numpy.all(numpoly.polynomial(
        numpy.array([[1, 2], [3, 4]])) == [[1, 2], [3, 4]])


def test_symbols():
    assert numpoly.symbols() == X
    assert numpoly.symbols().shape == ()
    assert numpoly.symbols("q0") == X
    assert numpoly.symbols("q0").shape == ()
    assert numpoly.symbols("q0,") == X
    assert numpoly.symbols("q0,").shape == (1,)
    assert numpoly.symbols("q0", asarray=True) == X
    assert numpoly.symbols("q0", asarray=True).shape == (1,)

    assert numpoly.symbols("q:1").names == ("q0",)
    assert numpoly.symbols("q1").names == ("q1",)



from setuptools import setup, Extension
from Cython.Build import cythonize
import numpy as np
import os

extensions = [
    Extension(
        name="numpoly.cfunctions." + item.replace(".pyx", ""),  # Name of the extension module
        sources=["./numpoly/cfunctions/" + item],  # Path to your `.pyx` file
        include_dirs=[np.get_include()],  # Include NumPy headers
    )
    for item in os.listdir("./numpoly/cfunctions/") if ".pyx" in item
]

# Setup configuration
setup(
    ext_modules=cythonize(extensions, compiler_directives={"language_level": 3}, verbose=True),  # Compile the Cython files
    include_dirs=[np.get_include()],  # Include NumPy headers
)



"""Split an array into multiple sub-arrays."""

from __future__ import annotations
from typing import List

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.split)
def split(
    ary: PolyLike,
    indices_or_sections: numpy.typing.ArrayLike,
    axis: int = 0,
) -> List[ndpoly]:
    """
    Split an array into multiple sub-arrays.

    Args:
        ary:
            Array to be divided into sub-arrays.
        indices_or_sections:
            If `indices_or_sections` is an integer, N, the array will be
            divided into N equal arrays along `axis`.  If such a split is not
            possible, an error is raised.

            If `indices_or_sections` is a 1-D array of sorted integers, the
            entries indicate where along `axis` the array is split.  For
            example, ``[2, 3]`` would, for ``axis=0``, result in

            - ary[:2]
            - ary[2:3]
            - ary[3:]

            If an index exceeds the dimension of the array along `axis`, an
            empty sub-array is returned correspondingly.
        axis:
            The axis along which to split, default is 0.

    Return:
        A list of sub-arrays.

    Raise:
        ValueError:
            If `indices_or_sections` is given as an integer, but a split does
            not result in equal division.

    Note:
        array_split:
            Split an array into multiple sub-arrays of equal or near-equal
            size. Does not raise an exception if an equal division cannot be
            made.
        hsplit:
            Split array into multiple sub-arrays horizontally (column-wise).
        vsplit:
            Split array into multiple sub-arrays vertically (row wise).
        dsplit:
            Split array into multiple sub-arrays along the 3rd axis (depth).

    Example:
        >>> poly = numpoly.monomial(16).reshape(4, 4)
        >>> poly
        polynomial([[1, q0, q0**2, q0**3],
                    [q0**4, q0**5, q0**6, q0**7],
                    [q0**8, q0**9, q0**10, q0**11],
                    [q0**12, q0**13, q0**14, q0**15]])
        >>> part1, _ = numpoly.split(poly, 2, axis=0)
        >>> part1
        polynomial([[1, q0, q0**2, q0**3],
                    [q0**4, q0**5, q0**6, q0**7]])
        >>> part1, _ = numpoly.split(poly, 2, axis=1)
        >>> part1
        polynomial([[1, q0],
                    [q0**4, q0**5],
                    [q0**8, q0**9],
                    [q0**12, q0**13]])

    """
    ary = numpoly.aspolynomial(ary)
    results = numpy.split(
        ary.values, indices_or_sections=indices_or_sections, axis=axis
    )
    return [
        numpoly.aspolynomial(result, names=ary.indeterminants) for result in results
    ]



"""Return the roots of a polynomial."""

from __future__ import annotations
import numpy
import numpoly

from ..baseclass import PolyLike

# from ..dispatch import implements


# @implements(numpy.roots)
def roots(poly: PolyLike) -> numpy.ndarray:
    """
    Return the roots of a polynomial.

    Assumes the polynomial has a single dimension.

    Args:
        poly:
            Polynomial to take roots of, or if constant, the coefficients of
            said polynomial. This to be compatible with :func:`numpy.roots`.

    Return:
        An array containing the roots of the polynomial.

    Raise:
        ValueError:
            When `poly` cannot be converted to a rank-1 polynomial.

    Note:
        The algorithm relies on computing the eigenvalues of the companion
        matrix [1]_.

    .. [1] R. A. Horn & C. R. Johnson, *Matrix Analysis*.  Cambridge, UK:
        Cambridge University Press, 1999, pp. 146-7.

    Example:
        >>> q0 = numpoly.variable()
        >>> poly = 3.2*q0**2+2*q0+1
        >>> numpoly.roots(poly)
        array([-0.3125+0.46351241j, -0.3125-0.46351241j])
        >>> numpoly.roots([3.2, 2, 1])
        array([-0.3125+0.46351241j, -0.3125-0.46351241j])

    """
    # backwards compatibility
    poly = numpoly.aspolynomial(poly)
    if poly.isconstant():
        return numpy.roots(poly.tonumpy())
    # only rank-1
    poly = numpoly.clean_attributes(poly, retain_names=False)
    if len(poly.names) > 1:
        raise ValueError("polynomial is not of rank 1.")
    # align exponents to include all coefficients
    filled_basis = poly.indeterminants ** numpy.arange(
        numpoly.lead_exponent(poly).item(), dtype=int
    )
    _, poly = numpoly.align_exponents(filled_basis, poly)
    # pass coefficients to numpy
    return numpy.roots(poly.coefficients[::-1])



"""Return a scalar type which is common to the input arrays."""

from __future__ import annotations
import numpy
import numpoly

from ..baseclass import PolyLike
from ..dispatch import implements


@implements(numpy.common_type)
def common_type(*arrays: PolyLike) -> numpy.dtype:
    """
    Return a scalar type which is common to the input arrays.

    The return type will always be an inexact (i.e. floating point) scalar
    type, even if all the arrays are integer arrays. If one of the inputs is an
    integer array, the minimum precision type that is returned is a 64-bit
    floating point dtype.

    All input arrays except int64 and uint64 can be safely cast to the
    returned dtype without loss of information.

    Args:
        arrays:
            Input arrays.

    Return:
        Data type code.

    Example:
        >>> scalar = numpy.array(2, dtype=numpy.float32)
        >>> numpoly.common_type(scalar) is numpy.float32
        True
        >>> q0 = numpoly.variable()
        >>> numpoly.common_type(q0) is numpy.float64
        True
        >>> numpoly.common_type(q0, 1j) is numpy.complex128
        True

    """
    arrays_ = [numpoly.aspolynomial(array) for array in arrays]
    arrays_ = [array.values[array.keys[0]] for array in arrays_]
    return numpy.common_type(*arrays_)



"""Compute the truth value of x1 AND x2 element-wise."""

from __future__ import annotations
from typing import Any, Optional, Sequence, Union

import numpy
import numpoly

from ..baseclass import PolyLike
from ..dispatch import implements


@implements(numpy.logical_and)
def logical_and(
    x1: PolyLike,
    x2: PolyLike,
    out: Optional[numpy.ndarray] = None,
    where: Union[bool, Sequence[bool]] = True,
    **kwargs: Any,
) -> numpy.ndarray:
    """
    Compute the truth value of x1 AND x2 element-wise.

    Args:
        x1, x2
            Input arrays. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Return:
        Boolean result of the logical OR operation applied to the elements of
        `x1` and `x2`; the shape is determined by broadcasting. This is a
        scalar if both `x1` and `x2` are scalars.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.logical_and(q0, 0)
        np.False_
        >>> numpoly.logical_and([q0, False], [q0, q1])
        array([ True, False])
        >>> const = numpy.arange(5)
        >>> numpoly.logical_and(const > 1, const < 4)
        array([False, False,  True,  True, False])

    """
    x1 = numpoly.aspolynomial(x1)
    x2 = numpoly.aspolynomial(x2)
    coefficients1 = numpy.any(numpy.asarray(x1.coefficients), 0)
    coefficients2 = numpy.any(numpy.asarray(x2.coefficients), 0)
    where_ = numpy.asarray(where)
    return numpy.logical_and(
        coefficients1, coefficients2, out=out, where=where_, **kwargs
    )



"""Create a numerical proxy for a polynomial to allow compare."""

from __future__ import annotations

import numpy
import numpoly

from ..baseclass import PolyLike


def sortable_proxy(
    poly: PolyLike,
    graded: bool = False,
    reverse: bool = False,
) -> numpy.ndarray:
    """
    Create a numerical proxy for a polynomial to allow compare.

    As polynomials are not inherently sortable, values are sorted using the
    highest `lexicographical` ordering. Between the values that have the same
    highest ordering, the elements are sorted using the coefficients. This also
    ensures that the method behaves as expected with ``numpy.ndarray``.

    Args:
        poly:
            Polynomial to convert into something sortable.
        graded:
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``q0**2*q1**2*q2**2`` has an exponent sum of 6, and will
            therefore be consider larger than both ``q0**3*q1*q2``,
            ``q0*q1**3*q2`` and ``q0*q1*z**3``.
        reverse:
            Reverses lexicographical sorting meaning that ``q0*q1**3`` is
            considered bigger than ``q0**3*q1``, instead of the opposite.

    Return:
        Integer array where ``a > b`` is retained for the giving rule of
        ``ordering``.

    Example:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial(
        ...     [q0**2, 2*q0, 3*q1, 4*q0, 5])
        >>> numpoly.sortable_proxy(poly)
        array([3, 1, 4, 2, 0])
        >>> numpoly.sortable_proxy(poly, reverse=True)
        array([4, 2, 1, 3, 0])
        >>> numpoly.sortable_proxy([8, 4, 10, -100])
        array([2, 1, 3, 0])
        >>> numpoly.sortable_proxy([[8, 4], [10, -100]])
        array([[2, 1],
               [3, 0]])

    """
    poly = numpoly.aspolynomial(poly)
    coefficients = poly.coefficients
    proxy = numpy.tile(-1, poly.shape)
    largest = numpoly.lead_exponent(poly, graded=graded, reverse=reverse)

    for idx in numpoly.glexsort(poly.exponents.T, graded=graded, reverse=reverse):
        indices = numpy.all(largest == poly.exponents[idx], axis=-1)
        values = numpy.argsort(coefficients[idx][indices])
        proxy[indices] = numpy.argsort(values) + numpy.max(proxy) + 1

    proxy = numpy.argsort(numpy.argsort(proxy.ravel())).reshape(proxy.shape)
    return proxy



"""Decompose a polynomial to component form."""

from __future__ import annotations
import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike


def decompose(poly: PolyLike) -> ndpoly:
    """
    Decompose a polynomial to component form.

    In array missing values are padded with 0 to make decomposition compatible
    with ``chaospy.sum(output, 0)``.

    Args:
        poly:
            Polynomial to decompose.

    Return:
        Decomposed polynomial with ``poly.shape==(M,)+output.shape``,
        where ``M`` is the number of components in `poly`.

    Example:
        >>> q0 = numpoly.variable()
        >>> poly = numpoly.polynomial([q0**2-1, 2])
        >>> poly
        polynomial([q0**2-1, 2])
        >>> numpoly.decompose(poly)
        polynomial([[-1, 2],
                    [q0**2, 0]])
        >>> numpoly.sum(numpoly.decompose(poly), 0)
        polynomial([q0**2-1, 2])

    """
    poly = numpoly.aspolynomial(poly)
    return numpoly.concatenate(
        [
            numpoly.construct.polynomial_from_attributes(
                exponents=[expon],
                coefficients=[numpy.asarray(poly.values[key])],
                names=poly.indeterminants,
                retain_coefficients=True,
                retain_names=True,
            )[numpy.newaxis]
            for key, expon in zip(poly.keys, poly.exponents)
        ]
    )



"""Split an array into multiple sub-arrays horizontally (column-wise)."""

from __future__ import annotations
from typing import Sequence

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.hsplit)
def hsplit(
    ary: PolyLike,
    indices_or_sections: numpy.typing.ArrayLike,
) -> Sequence[ndpoly]:
    """
    Split an array into multiple sub-arrays horizontally (column-wise).

    Please refer to the `split` documentation.  `hsplit` is equivalent to
    `split` with ``axis=1``, the array is always split along the second axis
    regardless of the array dimension.

    Note:
        split : Split an array into multiple sub-arrays of equal size.

    Example:
        >>> poly = numpoly.monomial(8).reshape(2, 4)
        >>> poly
        polynomial([[1, q0, q0**2, q0**3],
                    [q0**4, q0**5, q0**6, q0**7]])
        >>> part1, part2 = numpoly.hsplit(poly, 2)
        >>> part1
        polynomial([[1, q0],
                    [q0**4, q0**5]])
        >>> part2
        polynomial([[q0**2, q0**3],
                    [q0**6, q0**7]])
        >>> part1, part2, part3 = numpoly.hsplit(poly, [1, 2])
        >>> part1
        polynomial([[1],
                    [q0**4]])
        >>> part3
        polynomial([[q0**2, q0**3],
                    [q0**6, q0**7]])

    """
    ary = numpoly.aspolynomial(ary)
    results = numpy.hsplit(ary.values, indices_or_sections=indices_or_sections)
    return [
        numpoly.polynomial(result, names=ary.indeterminants, allocation=ary.allocation)
        for result in results
    ]



"""Monomial construction."""

from __future__ import annotations
from typing import Optional

import numpy
import numpy.typing

import numpoly
from ..baseclass import ndpoly


def monomial(
    start: numpy.typing.ArrayLike,
    stop: Optional[numpy.typing.ArrayLike] = None,
    dimensions: int = 1,
    cross_truncation: numpy.typing.ArrayLike = 1.0,
    graded: bool = False,
    reverse: bool = False,
    allocation: Optional[int] = None,
) -> ndpoly:
    """
    Create an polynomial monomial expansion.

    Args:
        start:
            The minimum polynomial to include. If int is provided, set as
            lowest total order. If array of int, set as lower order along each
            indeterminant.
        stop:
            The maximum shape included. If omitted:
            ``stop <- start; start <- 0`` If int is provided, set as largest
            total order. If array of int, set as largest order along each
            indeterminant.
        dimensions:
            The indeterminants names used to create the monomials expansion.
            If int provided, set the number of names to be used.
        cross_truncation:
            The hyperbolic cross truncation scheme to reduce the number of
            terms in expansion. In other words, it sets the :math:`q` in the
            :math:`L_q`-norm.
        graded:
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``q0**2*q1**2*q2**2`` has an exponent sum of 6, and will
            therefore be consider larger than both ``q0**3*q1*q2``,
            ``q0*q1**2*q2`` and ``q0*q1*q2**2``, which all have exponent
            sum of 5.
        reverse:
            Reverse lexicographical sorting meaning that ``q0*q1**3`` is
            considered bigger than ``q0**3*q1``, instead of the opposite.
        allocation:
            The maximum number of polynomial exponents. If omitted, use
            length of exponents for allocation.

    Return:
        Monomial expansion.

    Example:
        >>> numpoly.monomial(4)
        polynomial([1, q0, q0**2, q0**3])
        >>> numpoly.monomial(start=4, stop=5, dimensions=2,
        ...                  graded=True, reverse=True)
        polynomial([q1**4, q0*q1**3, q0**2*q1**2, q0**3*q1, q0**4])
        >>> numpoly.monomial(2, [3, 4], graded=True)
        polynomial([q0**2, q0*q1, q1**2, q1**3])
        >>> numpoly.monomial(
        ...     start=0,
        ...     stop=4,
        ...     dimensions=("q2", "q4"),
        ...     cross_truncation=0.5,
        ...     graded=True,
        ...     reverse=True,
        ... )
        polynomial([1, q4, q2, q4**2, q2**2, q4**3, q2**3])

    """
    if stop is None:
        start, stop = numpy.array(0), start

    start = numpy.array(start, dtype=int)
    stop = numpy.array(stop, dtype=int)
    if isinstance(dimensions, str):
        names, dimensions = (dimensions,), 1
    elif isinstance(dimensions, int):
        dimensions = max(start.size, stop.size, dimensions)
        names = numpoly.variable(dimensions).names
    elif dimensions is None:
        dimensions = max(start.size, stop.size)
        names = numpoly.variable(dimensions).names
    else:
        names, dimensions = dimensions, len(dimensions)

    indices = numpoly.glexindex(
        start=start,
        stop=stop,
        dimensions=dimensions,
        graded=graded,
        reverse=reverse,
        cross_truncation=cross_truncation,
    )
    poly = numpoly.ndpoly(
        exponents=indices,
        shape=(len(indices),),
        names=names,
        allocation=allocation,
    )
    for coeff, key in zip(numpy.eye(len(indices), dtype=int), poly.keys):
        numpy.ndarray.__setitem__(poly, key, coeff)
    return poly



"""Stack arrays in sequence horizontally (column wise)."""

from __future__ import annotations
from typing import Sequence

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.hstack)
def hstack(tup: Sequence[PolyLike]) -> ndpoly:
    """
    Stack arrays in sequence horizontally (column wise).

    This is equivalent to concatenation along the second axis, except for 1-D
    arrays where it concatenates along the first axis. Rebuilds arrays divided
    by `hsplit`.

    This function makes most sense for arrays with up to 3 dimensions. For
    instance, for pixel-data with a height (first axis), width (second axis),
    and r/g/b channels (third axis). The functions `concatenate`, `stack` and
    `block` provide more general stacking and concatenation operations.

    Args:
        tup:
            The arrays must have the same shape along all but the second axis,
            except 1-D arrays which can be any length.

    Return:
        The array formed by stacking the given arrays.

    Example:
        >>> poly1 = numpoly.variable(3)
        >>> const1 = numpoly.polynomial([1, 2, 3])
        >>> numpoly.hstack([poly1, const1])
        polynomial([q0, q1, q2, 1, 2, 3])
        >>> const2 = numpoly.polynomial([[1], [2], [3]])
        >>> poly2 = poly1.reshape(3, 1)
        >>> numpoly.hstack([const2, poly2])
        polynomial([[1, q0],
                    [2, q1],
                    [3, q2]])

    """
    arrays = numpoly.align_exponents(*tup)
    coefficients = [
        numpy.hstack([array.values[key] for array in arrays]) for key in arrays[0].keys
    ]
    return numpoly.polynomial_from_attributes(
        exponents=arrays[0].exponents,
        coefficients=coefficients,
        names=arrays[0].names,
        dtype=coefficients[0].dtype,
    )



"""Test element-wise for finiteness (not infinity or not Not a Number)."""

from __future__ import annotations
from typing import Any, Optional

import numpy
import numpy.typing

from ..baseclass import PolyLike
from ..dispatch import implements, simple_dispatch


@implements(numpy.isfinite)
def isfinite(
    x: PolyLike,
    out: Optional[numpy.ndarray] = None,
    where: numpy.typing.ArrayLike = True,
    **kwargs: Any,
) -> numpy.ndarray:
    """
    Test element-wise for finiteness (not infinity or not Not a Number).

    The result is returned as a boolean array.

    Args:
        x:
            Input values.
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value.
            Note that if an uninitialized `out` array is created via the
            default ``out=None``, locations within it where the condition is
            False will remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Return:
        True where ``x`` is not positive infinity, negative infinity, or NaN;
        false otherwise. This is a scalar if `x` is a scalar.

    Note:
        Not a Number, positive infinity and negative infinity are considered to
        be non-finite.

    Example:
        >>> numpoly.isfinite(1)
        np.True_
        >>> numpoly.isfinite(0)
        np.True_
        >>> numpoly.isfinite(numpy.nan*numpoly.variable())
        np.False_
        >>> numpoly.isfinite(numpy.inf)
        np.False_
        >>> numpoly.isfinite(-numpy.inf)
        np.False_
        >>> numpoly.isfinite([numpy.log(-1.), 1., numpy.log(0)])
        array([False,  True, False])

    """
    out_ = simple_dispatch(
        numpy_func=numpy.isfinite, inputs=(x,), where=where, **kwargs
    )
    if out is None:
        out_ = numpy.all(numpy.asarray(out_.coefficients), axis=0)
    else:
        out_ = numpy.all(numpy.asarray(out_.coefficients), out=out[0], axis=0)
    return out_



"""Split an array into multiple sub-arrays."""

from __future__ import annotations
from typing import List

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.array_split)
def array_split(
    ary: PolyLike,
    indices_or_sections: numpy.typing.ArrayLike,
    axis: int = 0,
) -> List[ndpoly]:
    """
    Split an array into multiple sub-arrays.

    Please refer to the ``split`` documentation.  The only difference between
    these functions is that ``array_split`` allows `indices_or_sections` to be
    an integer that does *not* equally divide the axis. For an array of length
    l that should be split into n sections, it returns l % n sub-arrays of size
    l//n + 1 and the rest of size l//n.

    Note:
        split : Split an array into multiple sub-arrays of equal size.

    Example:
        >>> poly = numpoly.monomial(8).reshape(2, 4)
        >>> poly
        polynomial([[1, q0, q0**2, q0**3],
                    [q0**4, q0**5, q0**6, q0**7]])
        >>> parts = numpoly.array_split(poly, 3, axis=1)
        >>> part1, part2, part3 = parts
        >>> part1
        polynomial([[1, q0],
                    [q0**4, q0**5]])
        >>> part2
        polynomial([[q0**2],
                    [q0**6]])
        >>> part3
        polynomial([[q0**3],
                    [q0**7]])

    """
    ary = numpoly.aspolynomial(ary)
    results = numpy.array_split(
        ary.values, indices_or_sections=indices_or_sections, axis=axis
    )
    return [
        numpoly.aspolynomial(result, names=ary.indeterminants) for result in results
    ]
